# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Tests for SequentialDecomposition
#

import pyomo.common.unittest as unittest

from pyomo.common.dependencies import numpy_available, networkx_available
from pyomo.environ import (
    SolverFactory,
    value,
    ConcreteModel,
    Set,
    Block,
    Var,
    TransformationFactory,
    Reference,
    Constraint,
)
from pyomo.network import Port, SequentialDecomposition, Arc
from pyomo.gdp.tests.models import makeExpandedNetworkDisjunction
from types import MethodType

import_available = numpy_available and networkx_available

glpk_available = SolverFactory('glpk').available(exception_flag=False)
ipopt_available = SolverFactory('ipopt').available(exception_flag=False)


@unittest.skipIf(not import_available, "numpy or networkx not available")
class TestSequentialDecomposition(unittest.TestCase):
    def is_converged(self, arc, rel=False, tol=1.0e-5):
        eblock = arc.expanded_block
        for name in arc.src.vars:
            if arc.src.vars[name].is_indexed():
                for i in arc.src.vars[name]:
                    sval = dval = None
                    if arc.src.is_extensive(name):
                        evar = eblock.component(name)
                        if evar is not None:
                            sf = eblock.component("splitfrac")
                            if sf is not None:
                                sval = value(arc.src.vars[name][i] * sf)
                            dval = value(evar[i])
                    if sval is None:
                        sval = value(arc.src.vars[name][i])
                    if dval is None:
                        dval = value(arc.dest.vars[name][i])
                    diff = abs(sval - dval)
                    if rel:
                        diff = diff / sval
                    if diff > tol:
                        return False
            else:
                sval = dval = None
                if arc.src.is_extensive(name):
                    evar = eblock.component(name)
                    if evar is not None:
                        sf = eblock.component("splitfrac")
                        if sf is not None:
                            sval = value(arc.src.vars[name] * sf)
                        dval = value(evar)
                if sval is None:
                    sval = value(arc.src.vars[name])
                if dval is None:
                    dval = value(arc.dest.vars[name])
                diff = abs(sval - dval)
                if rel:
                    diff = diff / sval
                if diff > tol:
                    return False

        return True

    def intensive_equal(self, port, tol=1.0e-5, **kwds):
        for name in kwds:
            if port.vars[name].is_indexed():
                for i in kwds[name]:
                    if abs(value(port.vars[name][i] - kwds[name][i])) > tol:
                        return False
                    if abs(value(port.vars[name][i] - kwds[name][i])) > tol:
                        return False
            else:
                if abs(value(port.vars[name] - kwds[name])) > tol:
                    return False
                if abs(value(port.vars[name] - kwds[name])) > tol:
                    return False

        return True

    def simple_recycle_model(self):
        m = ConcreteModel()
        m.comps = Set(initialize=["A", "B", "C"])

        # Feed
        m.feed = Block()

        m.feed.flow_out = Var(m.comps)
        m.feed.temperature_out = Var()
        m.feed.pressure_out = Var()

        m.feed.expr_var_idx_out = Var(m.comps)

        @m.feed.Expression(m.comps)
        def expr_idx_out(b, i):
            return -b.expr_var_idx_out[i]

        m.feed.expr_var_out = Var()
        m.feed.expr_out = -m.feed.expr_var_out

        @m.feed.Port()
        def outlet(b):
            return dict(
                flow=b.flow_out,
                temperature=b.temperature_out,
                pressure=b.pressure_out,
                expr_idx=b.expr_idx_out,
                expr=b.expr_out,
            )

        def initialize_feed(self):
            pass

        m.feed.initialize = MethodType(initialize_feed, m.feed)

        # Mixer
        m.mixer = Block()

        m.mixer.flow_in_side_1 = Var(m.comps)
        m.mixer.temperature_in_side_1 = Var()
        m.mixer.pressure_in_side_1 = Var()

        m.mixer.expr_var_idx_in_side_1 = Var(m.comps)

        @m.mixer.Expression(m.comps)
        def expr_idx_in_side_1(b, i):
            return -b.expr_var_idx_in_side_1[i]

        m.mixer.expr_var_in_side_1 = Var()
        m.mixer.expr_in_side_1 = -m.mixer.expr_var_in_side_1

        m.mixer.flow_in_side_2 = Var(m.comps)
        m.mixer.temperature_in_side_2 = Var()
        m.mixer.pressure_in_side_2 = Var()

        m.mixer.expr_var_idx_in_side_2 = Var(m.comps)

        @m.mixer.Expression(m.comps)
        def expr_idx_in_side_2(b, i):
            return -b.expr_var_idx_in_side_2[i]

        m.mixer.expr_var_in_side_2 = Var()
        m.mixer.expr_in_side_2 = -m.mixer.expr_var_in_side_2

        m.mixer.flow_out = Var(m.comps)
        m.mixer.temperature_out = Var()
        m.mixer.pressure_out = Var()

        m.mixer.expr_var_idx_out = Var(m.comps)

        @m.mixer.Expression(m.comps)
        def expr_idx_out(b, i):
            return -b.expr_var_idx_out[i]

        m.mixer.expr_var_out = Var()
        m.mixer.expr_out = -m.mixer.expr_var_out

        @m.mixer.Port()
        def inlet_side_1(b):
            return dict(
                flow=b.flow_in_side_1,
                temperature=b.temperature_in_side_1,
                pressure=b.pressure_in_side_1,
                expr_idx=b.expr_idx_in_side_1,
                expr=b.expr_in_side_1,
            )

        @m.mixer.Port()
        def inlet_side_2(b):
            return dict(
                flow=b.flow_in_side_2,
                temperature=b.temperature_in_side_2,
                pressure=b.pressure_in_side_2,
                expr_idx=b.expr_idx_in_side_2,
                expr=b.expr_in_side_2,
            )

        @m.mixer.Port()
        def outlet(b):
            return dict(
                flow=b.flow_out,
                temperature=b.temperature_out,
                pressure=b.pressure_out,
                expr_idx=b.expr_idx_out,
                expr=b.expr_out,
            )

        def initialize_mixer(self):
            for i in self.flow_out:
                self.flow_out[i].value = value(
                    self.flow_in_side_1[i] + self.flow_in_side_2[i]
                )
            for i in self.expr_var_idx_out:
                self.expr_var_idx_out[i].value = value(
                    self.expr_var_idx_in_side_1[i] + self.expr_var_idx_in_side_2[i]
                )
            self.expr_var_out.value = value(
                self.expr_var_in_side_1 + self.expr_var_in_side_2
            )
            assert self.temperature_in_side_1.value == self.temperature_in_side_2.value
            self.temperature_out.value = value(self.temperature_in_side_1)
            assert self.pressure_in_side_1.value == self.pressure_in_side_2.value
            self.pressure_out.value = value(self.pressure_in_side_1)

        m.mixer.initialize = MethodType(initialize_mixer, m.mixer)

        # Pass through
        m.unit = Block()

        m.unit.flow_in = Var(m.comps)
        m.unit.temperature_in = Var()
        m.unit.pressure_in = Var()

        m.unit.expr_var_idx_in = Var(m.comps)

        @m.unit.Expression(m.comps)
        def expr_idx_in(b, i):
            return -b.expr_var_idx_in[i]

        m.unit.expr_var_in = Var()
        m.unit.expr_in = -m.unit.expr_var_in

        m.unit.flow_out = Var(m.comps)
        m.unit.temperature_out = Var()
        m.unit.pressure_out = Var()

        m.unit.expr_var_idx_out = Var(m.comps)

        @m.unit.Expression(m.comps)
        def expr_idx_out(b, i):
            return -b.expr_var_idx_out[i]

        m.unit.expr_var_out = Var()
        m.unit.expr_out = -m.unit.expr_var_out

        @m.unit.Port()
        def inlet(b):
            return dict(
                flow=b.flow_in,
                temperature=b.temperature_in,
                pressure=b.pressure_in,
                expr_idx=b.expr_idx_in,
                expr=b.expr_in,
            )

        @m.unit.Port()
        def outlet(b):
            return dict(
                flow=b.flow_out,
                temperature=b.temperature_out,
                pressure=b.pressure_out,
                expr_idx=b.expr_idx_out,
                expr=b.expr_out,
            )

        def initialize_unit(self):
            for i in self.flow_out:
                self.flow_out[i].value = value(self.flow_in[i])
            for i in self.expr_var_idx_out:
                self.expr_var_idx_out[i].value = value(self.expr_var_idx_in[i])
            self.expr_var_out.value = value(self.expr_var_in)
            self.temperature_out.value = value(self.temperature_in)
            self.pressure_out.value = value(self.pressure_in)

        m.unit.initialize = MethodType(initialize_unit, m.unit)

        # Splitter
        m.splitter = Block()

        @m.splitter.Block(m.comps)
        def flow_in(b, i):
            b.flow = Var()

        m.splitter.temperature_in = Var()
        m.splitter.pressure_in = Var()

        m.splitter.expr_var_idx_in = Var(m.comps)

        @m.splitter.Expression(m.comps)
        def expr_idx_in(b, i):
            return -b.expr_var_idx_in[i]

        m.splitter.expr_var_in = Var()
        m.splitter.expr_in = -m.splitter.expr_var_in

        m.splitter.flow_out_side_1 = Var(m.comps)
        m.splitter.temperature_out_side_1 = Var()
        m.splitter.pressure_out_side_1 = Var()

        m.splitter.expr_var_idx_out_side_1 = Var(m.comps)

        @m.splitter.Expression(m.comps)
        def expr_idx_out_side_1(b, i):
            return -b.expr_var_idx_out_side_1[i]

        m.splitter.expr_var_out_side_1 = Var()
        m.splitter.expr_out_side_1 = -m.splitter.expr_var_out_side_1

        m.splitter.flow_out_side_2 = Var(m.comps)
        m.splitter.temperature_out_side_2 = Var()
        m.splitter.pressure_out_side_2 = Var()

        m.splitter.expr_var_idx_out_side_2 = Var(m.comps)

        @m.splitter.Expression(m.comps)
        def expr_idx_out_side_2(b, i):
            return -b.expr_var_idx_out_side_2[i]

        m.splitter.expr_var_out_side_2 = Var()
        m.splitter.expr_out_side_2 = -m.splitter.expr_var_out_side_2

        @m.splitter.Port()
        def inlet(b):
            return dict(
                flow=Reference(b.flow_in[:].flow),
                temperature=b.temperature_in,
                pressure=b.pressure_in,
                expr_idx=b.expr_idx_in,
                expr=b.expr_in,
            )

        @m.splitter.Port()
        def outlet_side_1(b):
            return dict(
                flow=b.flow_out_side_1,
                temperature=b.temperature_out_side_1,
                pressure=b.pressure_out_side_1,
                expr_idx=b.expr_idx_out_side_1,
                expr=b.expr_out_side_1,
            )

        @m.splitter.Port()
        def outlet_side_2(b):
            return dict(
                flow=b.flow_out_side_2,
                temperature=b.temperature_out_side_2,
                pressure=b.pressure_out_side_2,
                expr_idx=b.expr_idx_out_side_2,
                expr=b.expr_out_side_2,
            )

        def initialize_splitter(self):
            recycle = 0.1
            prod = 1 - recycle
            for i in self.flow_in:
                self.flow_out_side_1[i].value = prod * value(self.flow_in[i].flow)
                self.flow_out_side_2[i].value = recycle * value(self.flow_in[i].flow)
            for i in self.expr_var_idx_in:
                self.expr_var_idx_out_side_1[i].value = prod * value(
                    self.expr_var_idx_in[i]
                )
                self.expr_var_idx_out_side_2[i].value = recycle * value(
                    self.expr_var_idx_in[i]
                )
            self.expr_var_out_side_1.value = prod * value(self.expr_var_in)
            self.expr_var_out_side_2.value = recycle * value(self.expr_var_in)
            self.temperature_out_side_1.value = value(self.temperature_in)
            self.temperature_out_side_2.value = value(self.temperature_in)
            self.pressure_out_side_1.value = value(self.pressure_in)
            self.pressure_out_side_2.value = value(self.pressure_in)

        m.splitter.initialize = MethodType(initialize_splitter, m.splitter)

        # Prod
        m.prod = Block()

        m.prod.flow_in = Var(m.comps)
        m.prod.temperature_in = Var()
        m.prod.pressure_in = Var()

        m.prod.actual_var_idx_in = Var(m.comps)
        m.prod.actual_var_in = Var()

        @m.prod.Port()
        def inlet(b):
            return dict(
                flow=b.flow_in,
                temperature=b.temperature_in,
                pressure=b.pressure_in,
                expr_idx=b.actual_var_idx_in,
                expr=b.actual_var_in,
            )

        def initialize_prod(self):
            pass

        m.prod.initialize = MethodType(initialize_prod, m.prod)

        # Arcs
        @m.Arc(directed=True)
        def stream_feed_to_mixer(m):
            return (m.feed.outlet, m.mixer.inlet_side_1)

        @m.Arc(directed=True)
        def stream_mixer_to_unit(m):
            return (m.mixer.outlet, m.unit.inlet)

        @m.Arc(directed=True)
        def stream_unit_to_splitter(m):
            return (m.unit.outlet, m.splitter.inlet)

        @m.Arc(directed=True)
        def stream_splitter_to_mixer(m):
            return (m.splitter.outlet_side_2, m.mixer.inlet_side_2)

        @m.Arc(directed=True)
        def stream_splitter_to_prod(m):
            return (m.splitter.outlet_side_1, m.prod.inlet)

        # Expand Arcs
        TransformationFactory("network.expand_arcs").apply_to(m)

        # Fix Feed
        m.feed.flow_out['A'].fix(100)
        m.feed.flow_out['B'].fix(200)
        m.feed.flow_out['C'].fix(300)
        m.feed.expr_var_idx_out['A'].fix(10)
        m.feed.expr_var_idx_out['B'].fix(20)
        m.feed.expr_var_idx_out['C'].fix(30)
        m.feed.expr_var_out.fix(40)
        m.feed.temperature_out.fix(450)
        m.feed.pressure_out.fix(128)

        return m

    def simple_recycle_run(self, tear_method, tol_type):
        rel = tol_type == "rel"
        m = self.simple_recycle_model()

        def function(unit):
            unit.initialize()

        seq = SequentialDecomposition(tear_method=tear_method, tol_type=tol_type)
        tset = [m.stream_splitter_to_mixer]
        seq.set_tear_set(tset)
        splitter_to_mixer_guess = {
            "flow": {"A": 0, "B": 0, "C": 0},
            "temperature": 450,
            "pressure": 128,
        }
        seq.set_guesses_for(m.mixer.inlet_side_2, splitter_to_mixer_guess)
        # need to set guesses for expression members by initializing those vars
        m.mixer.expr_var_idx_in_side_2["A"] = 0
        m.mixer.expr_var_idx_in_side_2["B"] = 0
        m.mixer.expr_var_idx_in_side_2["C"] = 0
        m.mixer.expr_var_in_side_2 = 0
        seq.run(m, function)

        self.check_recycle_model(m, rel=rel)

    def extensive_recycle_model(self):
        def build_in_out(b):
            b.flow_in = Var(m.comps)
            b.mass_in = Var()
            b.temperature_in = Var()
            b.pressure_in = Var()

            b.expr_var_idx_in = Var(m.comps)

            @b.Expression(m.comps)
            def expr_idx_in(b, i):
                return -b.expr_var_idx_in[i]

            b.expr_var_in = Var()
            b.expr_in = -b.expr_var_in

            b.flow_out = Var(m.comps)
            b.mass_out = Var()
            b.temperature_out = Var()
            b.pressure_out = Var()

            b.expr_var_idx_out = Var(m.comps)

            @b.Expression(m.comps)
            def expr_idx_out(b, i):
                return -b.expr_var_idx_out[i]

            b.expr_var_out = Var()
            b.expr_out = -b.expr_var_out

            b.inlet = Port(rule=inlet)
            b.outlet = Port(rule=outlet)

            b.initialize = MethodType(initialize, b)

        def inlet(b):
            return dict(
                flow=(b.flow_in, Port.Extensive),
                mass=(b.mass_in, Port.Extensive),
                temperature=b.temperature_in,
                pressure=b.pressure_in,
                expr_idx=(b.expr_idx_in, Port.Extensive),
                expr=(b.expr_in, Port.Extensive),
            )

        def outlet(b):
            return dict(
                flow=(b.flow_out, Port.Extensive),
                mass=(b.mass_out, Port.Extensive),
                temperature=b.temperature_out,
                pressure=b.pressure_out,
                expr_idx=(b.expr_idx_out, Port.Extensive),
                expr=(b.expr_out, Port.Extensive),
            )

        def initialize(self):
            for i in self.flow_out:
                self.flow_out[i].value = value(self.flow_in[i])
            self.mass_out.value = value(self.mass_in)
            for i in self.expr_var_idx_out:
                self.expr_var_idx_out[i].value = value(self.expr_var_idx_in[i])
            self.expr_var_out.value = value(self.expr_var_in)
            self.temperature_out.value = value(self.temperature_in)
            self.pressure_out.value = value(self.pressure_in)

        def nop(self):
            pass

        m = ConcreteModel()
        m.comps = Set(initialize=["A", "B", "C"])

        # Feed
        m.feed = Block()

        m.feed.flow_out = Var(m.comps)
        m.feed.mass_out = Var()
        m.feed.temperature_out = Var()
        m.feed.pressure_out = Var()

        m.feed.expr_var_idx_out = Var(m.comps)

        @m.feed.Expression(m.comps)
        def expr_idx_out(b, i):
            return -b.expr_var_idx_out[i]

        m.feed.expr_var_out = Var()
        m.feed.expr_out = -m.feed.expr_var_out

        m.feed.outlet = Port(rule=outlet)

        m.feed.initialize = MethodType(nop, m.feed)

        # Mixer
        m.mixer = Block()
        build_in_out(m.mixer)

        # Pass through
        m.unit = Block()
        build_in_out(m.unit)

        # Splitter
        m.splitter = Block()
        build_in_out(m.splitter)

        # Prod
        m.prod = Block()

        m.prod.flow_in = Var(m.comps)
        m.prod.mass_in = Var()
        m.prod.temperature_in = Var()
        m.prod.pressure_in = Var()

        m.prod.actual_var_idx_in = Var(m.comps)
        m.prod.actual_var_in = Var()

        @m.prod.Port()
        def inlet(b):
            return dict(
                flow=(b.flow_in, Port.Extensive),
                mass=(b.mass_in, Port.Extensive),
                temperature=b.temperature_in,
                pressure=b.pressure_in,
                expr_idx=(b.actual_var_idx_in, Port.Extensive),
                expr=(b.actual_var_in, Port.Extensive),
            )

        m.prod.initialize = MethodType(nop, m.prod)

        # Arcs
        @m.Arc(directed=True)
        def stream_feed_to_mixer(m):
            return (m.feed.outlet, m.mixer.inlet)

        @m.Arc(directed=True)
        def stream_mixer_to_unit(m):
            return (m.mixer.outlet, m.unit.inlet)

        @m.Arc(directed=True)
        def stream_unit_to_splitter(m):
            return (m.unit.outlet, m.splitter.inlet)

        @m.Arc(directed=True)
        def stream_splitter_to_mixer(m):
            return (m.splitter.outlet, m.mixer.inlet)

        @m.Arc(directed=True)
        def stream_splitter_to_prod(m):
            return (m.splitter.outlet, m.prod.inlet)

        # Split Fraction
        rec = 0.1
        prod = 1 - rec
        m.splitter.outlet.set_split_fraction(m.stream_splitter_to_mixer, rec)
        m.splitter.outlet.set_split_fraction(m.stream_splitter_to_prod, prod)

        # Expand Arcs
        TransformationFactory("network.expand_arcs").apply_to(m)

        # Fix Feed
        m.feed.flow_out['A'].fix(100)
        m.feed.flow_out['B'].fix(200)
        m.feed.flow_out['C'].fix(300)
        m.feed.mass_out.fix(400)
        m.feed.expr_var_idx_out['A'].fix(10)
        m.feed.expr_var_idx_out['B'].fix(20)
        m.feed.expr_var_idx_out['C'].fix(30)
        m.feed.expr_var_out.fix(40)
        m.feed.temperature_out.fix(450)
        m.feed.pressure_out.fix(128)

        return m

    def extensive_recycle_run(self, tear_method, tol_type):
        rel = tol_type == "rel"
        m = self.extensive_recycle_model()

        def function(unit):
            unit.initialize()

        seq = SequentialDecomposition(tear_method=tear_method, tol_type=tol_type)
        tset = [m.stream_splitter_to_mixer]
        seq.set_tear_set(tset)
        splitter_to_mixer_guess = {
            "flow": {
                "A": [(m.stream_splitter_to_mixer, 0)],
                "B": [(m.stream_splitter_to_mixer, 0)],
                "C": [(m.stream_splitter_to_mixer, 0)],
            },
            "mass": [(m.stream_splitter_to_mixer, 0)],
            "expr_idx": {
                "A": [(m.stream_splitter_to_mixer, 0)],
                "B": [(m.stream_splitter_to_mixer, 0)],
                "C": [(m.stream_splitter_to_mixer, 0)],
            },
            "expr": [(m.stream_splitter_to_mixer, 0)],
            "temperature": 450,
            "pressure": 128,
        }
        seq.set_guesses_for(m.mixer.inlet, splitter_to_mixer_guess)
        seq.run(m, function)

        self.check_recycle_model(m, rel=rel)

        if rel:
            s = value(m.prod.inlet.mass)
            d = value(m.feed.outlet.mass)
            self.assertAlmostEqual((s - d) / s, 0, places=5)
        else:
            self.assertAlmostEqual(
                value(m.prod.inlet.mass), value(m.feed.outlet.mass), places=5
            )

    def check_recycle_model(self, m, rel=False):
        for arc in m.component_data_objects(Arc):
            self.assertTrue(self.is_converged(arc, rel=rel))

        for port in m.component_data_objects(Port):
            self.assertTrue(
                self.intensive_equal(
                    port,
                    temperature=value(m.feed.outlet.temperature),
                    pressure=value(m.feed.outlet.pressure),
                )
            )

        if rel:
            # in == out
            for i in m.feed.outlet.flow:
                s = value(m.prod.inlet.flow[i])
                d = value(m.feed.outlet.flow[i])
                self.assertAlmostEqual((s - d) / s, 0, places=5)
            for i in m.feed.outlet.expr_idx:
                s = value(m.prod.inlet.expr_idx[i])
                d = value(m.feed.outlet.expr_idx[i])
                self.assertAlmostEqual((s - d) / s, 0, places=5)
            s = value(m.prod.inlet.expr)
            d = value(m.feed.outlet.expr)
            self.assertAlmostEqual((s - d) / s, 0, places=5)

            # check the expressions work, should be negative in prod
            for i in m.feed.outlet.expr_idx:
                s = value(-m.prod.actual_var_idx_in[i])
                d = value(m.feed.expr_var_idx_out[i])
                self.assertAlmostEqual((s - d) / s, 0, places=5)
            s = value(-m.prod.actual_var_in)
            d = value(m.feed.expr_var_out)
            self.assertAlmostEqual((s - d) / s, 0, places=5)
        else:
            # in == out
            for i in m.feed.outlet.flow:
                self.assertAlmostEqual(
                    value(m.prod.inlet.flow[i]), value(m.feed.outlet.flow[i]), places=5
                )
            for i in m.feed.outlet.expr_idx:
                self.assertAlmostEqual(
                    value(m.prod.inlet.expr_idx[i]),
                    value(m.feed.outlet.expr_idx[i]),
                    places=5,
                )
            self.assertAlmostEqual(
                value(m.prod.inlet.expr), value(m.feed.outlet.expr), places=5
            )

            # check the expressions work, should be negative in prod
            for i in m.feed.outlet.expr_idx:
                self.assertAlmostEqual(
                    value(-m.prod.actual_var_idx_in[i]),
                    value(m.feed.expr_var_idx_out[i]),
                    places=5,
                )
            self.assertAlmostEqual(
                value(-m.prod.actual_var_in), value(m.feed.expr_var_out), places=5
            )

    def test_simple_recycle_direct_abs(self):
        self.simple_recycle_run(tear_method="Direct", tol_type="abs")

    def test_simple_recycle_wegstein_abs(self):
        self.simple_recycle_run(tear_method="Wegstein", tol_type="abs")

    def test_simple_recycle_direct_rel(self):
        self.simple_recycle_run(tear_method="Direct", tol_type="rel")

    def test_simple_recycle_wegstein_rel(self):
        self.simple_recycle_run(tear_method="Wegstein", tol_type="rel")

    def test_extensive_recycle_direct_abs(self):
        self.extensive_recycle_run(tear_method="Direct", tol_type="abs")

    def test_extensive_recycle_wegstein_abs(self):
        self.extensive_recycle_run(tear_method="Wegstein", tol_type="abs")

    def test_extensive_recycle_direct_rel(self):
        self.extensive_recycle_run(tear_method="Direct", tol_type="rel")

    def test_extensive_recycle_wegstein_rel(self):
        self.extensive_recycle_run(tear_method="Wegstein", tol_type="rel")

    @unittest.skipIf(not glpk_available, "GLPK solver not available")
    def test_tear_selection(self):
        m = self.simple_recycle_model()
        seq = SequentialDecomposition()
        G = seq.create_graph(m)

        heu_result = seq.select_tear_heuristic(G)
        self.assertEqual(heu_result[1], 1)
        self.assertEqual(heu_result[2], 1)

        all_tsets = []
        for tset in heu_result[0]:
            all_tsets.append(seq.indexes_to_arcs(G, tset))
        for arc in (
            m.stream_mixer_to_unit,
            m.stream_unit_to_splitter,
            m.stream_splitter_to_mixer,
        ):
            self.assertIn([arc], all_tsets)

        tset_mip = seq.tear_set_arcs(G, "mip", solver="glpk")
        self.assertIn(tset_mip, all_tsets)

        tset_heu = seq.tear_set_arcs(G, "heuristic")
        self.assertIn(tset_heu, all_tsets)

    @unittest.skipIf(not glpk_available, "GLPK solver not available")
    def test_select_tear_in_run(self):
        m = self.simple_recycle_model()

        def function(unit):
            unit.initialize()

        seq = SequentialDecomposition()
        tset = [m.stream_splitter_to_mixer]
        seq.set_tear_set(tset)
        splitter_to_mixer_guess = {
            "flow": {"A": 0, "B": 0, "C": 0},
            "temperature": 450,
            "pressure": 128,
        }
        seq.set_guesses_for(m.mixer.inlet_side_2, splitter_to_mixer_guess)
        # need to set guesses for expression members by initializing those vars
        m.mixer.expr_var_idx_in_side_2["A"] = 0
        m.mixer.expr_var_idx_in_side_2["B"] = 0
        m.mixer.expr_var_idx_in_side_2["C"] = 0
        m.mixer.expr_var_in_side_2 = 0
        seq.run(m, function)

        # we shouldn't need to know which streams are torn since everything
        # should already have values set so we don't need guesses, but we
        # just make sure it is able to select a tear set on its own
        seq = SequentialDecomposition(tear_solver="glpk", select_tear_method="mip")
        seq.run(m, function)
        self.check_recycle_model(m)

        seq = SequentialDecomposition(
            tear_solver="glpk", select_tear_method="heuristic"
        )
        seq.run(m, function)
        self.check_recycle_model(m)

    def _test_disjuncts(self, blue_on=True):
        m = makeExpandedNetworkDisjunction()
        if blue_on:
            m.blue.indicator_var.fix(1)
            m.orange.indicator_var.fix(0)
        else:
            m.blue.indicator_var.fix(0)
            m.orange.indicator_var.fix(1)
        TransformationFactory('gdp.fix_disjuncts').apply_to(m)

        def initializer(blk):
            for _ in blk.component_data_objects(Constraint, active=True):
                SolverFactory("ipopt").solve(blk)
                break

        seq = SequentialDecomposition(select_tear_method="heuristic", default_guess=0.5)
        seq.run(m, initializer)
        if blue_on:
            self.assertAlmostEqual(value(m.dest.x), 0.84)
        else:
            self.assertAlmostEqual(value(m.dest.x), 0.42)

    @unittest.skipIf(not ipopt_available, "ipopt solver not available")
    def test_fixed_disjuncts(self):
        self._test_disjuncts(True)
        self._test_disjuncts(False)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os
import types

try:
    import new

    new_available = True
except:
    new_available = False

import pyomo.common.unittest as unittest

from pyomo.solvers.tests.models.base import all_models
from pyomo.solvers.tests.testcases import generate_scenarios
from pyomo.common.log import LoggingIntercept

from io import StringIO

# The test directory
thisDir = os.path.dirname(os.path.abspath(__file__))

# Cleanup Expected Failure Results Files
_cleanup_expected_failures = True


#
# A function that returns a function that gets
# added to a test class.
#
def create_method(model, solver, io, test_case):
    is_expected_failure = test_case.status == 'expected failure'

    #
    # Create a function that executes the test
    #
    def failed_solve_test(self):
        # Create the model test class
        model_class = test_case.model()

        # Create the model instance
        model_class.generate_model(test_case.testcase.import_suffixes)
        model_class.warmstart_model()

        # solve
        load_solutions = True
        symbolic_labels = False
        out = StringIO()
        with LoggingIntercept(out, 'pyomo.solvers'):
            with LoggingIntercept(out, 'pyomo.opt'):
                opt, results = model_class.solve(
                    solver,
                    io,
                    test_case.testcase.io_options,
                    test_case.testcase.options,
                    symbolic_labels,
                    load_solutions,
                )
        model_class.post_solve_test_validation(self, results)
        if len(results.solution) == 0:
            self.assertIn("No solution is available", out.getvalue())
        else:
            # Note ASL solvers might still return a solution
            # file with garbage values in it for a failed solve
            self.assertEqual(len(results.solution), 1)

    # Skip this test if the status is 'skip'
    if test_case.status == 'skip':

        def return_test(self):
            return self.skipTest(test_case.msg)

    elif is_expected_failure:

        @unittest.expectedFailure
        def return_test(self):
            return failed_solve_test(self)

    else:
        # Return a normal test
        def return_test(self):
            return failed_solve_test(self)

    unittest.pytest.mark.solver(solver)(unittest.pytest.mark.writer(io)(return_test))
    return return_test


cls = None

#
# Create test driver classes for each test model
#
driver = {}
for model in all_models():
    # Get the test case for the model
    case = all_models(model)
    if case().solve_should_fail:
        # Create the test class
        name = "Test_%s" % model
        if new_available:
            cls = new.classobj(name, (unittest.TestCase,), {})
        else:
            cls = types.new_class(name, (unittest.TestCase,))
            cls.__module__ = __name__
        driver[model] = cls
        globals()[name] = cls

#
# Iterate through all test scenarios and add test methods
#
for key, value in generate_scenarios():
    model, solver, io = key

    if model in driver:
        cls = driver[model]
        # TODO: expand these tests to cover ASL models once
        #       a change in load_solutions behavior is
        #       propagated into that framework.
        if "_kernel" in cls.__name__:
            test_name = "test_" + solver + "_" + io
            test_method = create_method(model, solver, io, value)
            if test_method is not None:
                setattr(cls, test_name, test_method)
                test_method = None

# Reset the cls variable, since it contains a unittest.TestCase subclass.
# This prevents this class from being processed twice!
cls = None

if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
import pyomo
import inspect
import sys


class FlagType(type):
    """Metaclass to help generate "Flag Types".

    This is useful for defining "flag types" that are default arguments
    in functions so that the Sphinx-generated documentation is
    "cleaner".  These types are not constructable (attempts to construct
    the class return the class) and simplify the repr(type) and
    str(type).

    This metaclass redefines the ``str()`` and ``repr()`` of resulting
    classes.  The str() of the class returns only the class' ``__name__``,
    whereas the repr() returns either the qualified class name
    (``__qualname__``) if Sphinx has been imported, or else the
    fully-qualified class name (``__module__ + '.' + __qualname__``).

    """

    def __new__(mcs, name, bases, dct):
        # Ensure that attempts to construct instances of a Flag type
        # return the type.
        def __new_flag__(cls, *args, **kwargs):
            return cls

        dct["__new__"] = __new_flag__
        return type.__new__(mcs, name, bases, dct)

    def __repr__(cls):
        if building_documentation():
            return cls.__qualname__
        else:
            return cls.__module__ + "." + cls.__qualname__

    def __str__(cls):
        return cls.__name__


class NOTSET(object, metaclass=FlagType):
    """
    Class to be used to indicate that an optional argument
    was not specified, if `None` may be ambiguous. Usage:

    Examples
    --------
    >>> def foo(value=NOTSET):
    ...     if value is NOTSET:
    ...         pass  # no argument was provided to `value`

    """

    pass


def in_testing_environment(state=NOTSET):
    """Return True if we are currently running in a "testing" environment

    This currently includes if ``nose``, ``nose2``, or ``pytest`` are
    running (imported).

    Parameters
    ----------
    state : bool or None
        If provided, sets the current state of the testing environment
        (Setting to None reverts to the normal interrogation of
        ``sys.modules``)

    Returns
    -------
    bool

    """
    if state is not NOTSET:
        in_testing_environment.state = state
    if in_testing_environment.state is not None:
        return bool(in_testing_environment.state)
    return any(mod in sys.modules for mod in ('nose', 'nose2', 'pytest'))


in_testing_environment.state = None


def building_documentation(state=NOTSET):
    """True if we are building the Sphinx documentation

    We detect if we are building the documentation by looking if the
    ``sphinx`` or ``Sphinx`` modules are imported.

    Parameters
    ----------
    state : bool or None
        If provided, sets the current state of the building environment
        flag (Setting to None reverts to the normal interrogation of
        ``sys.modules``)

    Returns
    -------
    bool

    """
    if state is not NOTSET:
        building_documentation.state = state
    if building_documentation.state is not None:
        return bool(building_documentation.state)
    # Note that we previously detected if Sphinx was running by looking
    # for it in sys.modules.  That proved to be fragile as other
    # packages (notably sphinx-jinja2-compat) started causing Sphinx to
    # be imported immediately any time Python started.  We work around
    # this by having Pyomo's Sphinx conf.py add a flag to an
    # easy-to-find location.
    return bool(getattr(pyomo, '__sphinx_build__', False))


building_documentation.state = None


def serializing():
    """True if it looks like we are serializing objects

    This looks through the call stack and returns True if it finds a
    `dump` function anywhere in the call stack.  While not foolproof,
    this should reliably catch most serializers, including ``pickle``
    and `yaml``.

    """
    # Start by skipping this function
    frame = inspect.currentframe().f_back
    while frame is not None:
        if frame.f_code.co_name == 'dump':
            return True
        frame = frame.f_back
    return False



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Development of this module was conducted as part of the Institute for
# the Design of Advanced Energy Systems (IDAES) with support through the
# Simulation-Based Engineering, Crosscutting Research Program within the
# U.S. Department of Energy's Office of Fossil Energy and Carbon Management.
# ____________________________________________________________________________________



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.environ
from pyomo.core import *

# create the concrete model
model = ConcreteModel()

# set the data (native python data)
k1 = 5.0 / 6.0  # min^-1
k2 = 5.0 / 3.0  # min^-1
k3 = 1.0 / 6000.0  # m^3/(gmol min)
caf = 10000.0  # gmol/m^3

# create the variables
model.sv = Var(initialize=1.0, within=PositiveReals)
model.ca = Var(initialize=5000.0, bounds=(0, 10000))
model.cb = Var(initialize=2000.0, within=PositiveReals)
model.cc = Var(initialize=2000.0, within=PositiveReals)
model.cd = Var(initialize=1000.0, within=PositiveReals)

# create the objective
model.obj = Objective(expr=model.cb, sense=maximize)

# create the constraints
model.ca_bal = Constraint(
    expr=(
        0
        == model.sv * caf
        - model.sv * model.ca
        - k1 * model.ca
        - 2.0 * k3 * model.ca**2.0
    )
)

model.cb_bal = Constraint(
    expr=(0 == -model.sv * model.cb + k1 * model.ca - k2 * model.cb)
)

model.cc_bal = Constraint(expr=(0 == -model.sv * model.cc + k2 * model.cb))

model.cd_bal = Constraint(expr=(0 == -model.sv * model.cd + k3 * model.ca**2.0))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import re

from pyomo.common.deprecation import deprecated
from pyomo.core.base.componentuid import ComponentUID

# This module provides some basic functionality for generating labels
# from pyomo names, which often contain characters such as "[" and "]"
# (e.g., in my_var[1]).  These characters generally cause issues with
# optimization input file formats, e.g., CPLEX LP files.  The purpose of
# this module is to provide a simple remap function, that will satisfy
# broadly problematic symbols. if solver-specific remaps are required,
# they should be handled in the corresponding solver plugin.


class _CharMapper:
    def __init__(self, preserve, translate, other):
        """
        Arguments::
           preserve: a string of characters to preserve
           translate: a dict or key/value list of characters to translate
           other: the character to return for all characters not in
                  preserve or translate
        """
        self.table = {
            k if isinstance(k, int) else ord(k): v for k, v in dict(translate).items()
        }
        for c in preserve:
            _c = ord(c)
            if _c in self.table and self.table[_c] != c:
                raise RuntimeError(
                    "Duplicate character '%s' appears in both "
                    "translate table and preserve list" % (c,)
                )
            self.table[_c] = c
        self.other = other

    def __getitem__(self, c):
        # Most of the time c should be known.  For the rare cases we
        # encounter a new character, remember it by adding a new entry
        # into the translation table and return the default character
        try:
            return self.table[c]
        except:
            self.table[c] = self.other
            return self.other

    def make_table(self):
        return ''.join(self[i] for i in range(256))


_alpha = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJLKMNOPQRSTUVWXYZ'
_digit = '1234567890'
_cpxlp_translation_table = _CharMapper(
    preserve=_alpha + _digit + '()_', translate=zip('[]{}', '()()'), other='_'
).make_table()


def cpxlp_label_from_name(name):
    if name is None:
        raise RuntimeError(
            "Illegal name=None supplied to cpxlp_label_from_name function"
        )

    return str.translate(name, _cpxlp_translation_table)


_alphanum_translation_table = _CharMapper(
    preserve=_alpha + _digit + '_', translate={}, other='_'
).make_table()


def alphanum_label_from_name(name):
    if name is None:
        raise RuntimeError(
            "Illegal name=None supplied to alphanum_label_from_name function"
        )

    return str.translate(name, _alphanum_translation_table)


class CuidLabeler:
    def __call__(self, obj=None):
        return ComponentUID(obj)


class CounterLabeler:
    def __init__(self, start=0):
        self._id = start

    def __call__(self, obj=None):
        self._id += 1
        return self._id


class NumericLabeler:
    def __init__(self, prefix, start=0):
        self.id = start
        self.prefix = prefix

    def __call__(self, obj=None):
        self.id += 1
        return self.prefix + str(self.id)

    @deprecated(
        "The 'remove_obj' method is no longer "
        "necessary now that 'getname' does not "
        "support the use of a name buffer",
        version="6.4.1",
    )
    def remove_obj(self, obj):
        pass


#
# TODO: [JDS] I would like to rename TextLabeler to LPFileLabeler - as it
# generated LP-file-compliant labels - and make the CNameLabeler the
# TextLabeler.  This makes sense as the name() is the closest thing we
# have to a human-readable canonical text naming convention (the
# ComponentUID strings are actually unique, but not meant to be human
# readable).  Unfortunately, the TextLabeler is used all over the place
# (particularly PySP), and I don't know how much depends on the labels
# actually being LP-compliant.
#
class CNameLabeler:
    def __call__(self, obj):
        return obj.getname(True)


class LPFileLabeler:
    def __call__(self, obj):
        return cpxlp_label_from_name(obj.getname(True))

    @deprecated(
        "The 'remove_obj' method is no longer "
        "necessary now that 'getname' does not "
        "support the use of a name buffer",
        version="6.4.1",
    )
    def remove_obj(self, obj):
        pass


TextLabeler = LPFileLabeler


class AlphaNumericTextLabeler:
    def __call__(self, obj):
        return alphanum_label_from_name(obj.getname(True))


class NameLabeler:
    def __call__(self, obj):
        return obj.getname(True)


class ShortNameLabeler:
    def __init__(
        self,
        limit,
        suffix,
        start=0,
        labeler=None,
        prefix="",
        caseInsensitive=False,
        legalRegex=None,
    ):
        self.id = start
        self.prefix = prefix
        self.suffix = suffix
        self.limit = limit
        if labeler is not None:
            self.labeler = labeler
        else:
            self.labeler = AlphaNumericTextLabeler()
        self.known_labels = set()
        self.caseInsensitive = caseInsensitive
        if isinstance(legalRegex, str):
            self.legalRegex = re.compile(legalRegex)
        else:
            self.legalRegex = legalRegex

    def __call__(self, obj=None):
        lbl = self.labeler(obj)
        lbl_len = len(lbl)
        shorten = False
        if lbl_len > self.limit:
            shorten = True
        elif (
            lbl_len == self.limit
            and lbl.startswith(self.prefix)
            and lbl.endswith(self.suffix)
        ):
            shorten = True
        elif (lbl.upper() if self.caseInsensitive else lbl) in self.known_labels:
            shorten = True
        elif self.legalRegex and not self.legalRegex.match(lbl):
            shorten = True
        if shorten:
            self.id += 1
            suffix = "%s%d%s" % (self.suffix, self.id, self.suffix)
            tail = -self.limit + len(suffix) + len(self.prefix)
            if tail >= 0:
                raise RuntimeError(
                    "Too many identifiers.\n\t"
                    "The ShortNameLabeler cannot generate a guaranteed unique "
                    "label limited to %d characters" % (self.limit,)
                )
            lbl = self.prefix + lbl[tail:] + suffix
        if self.known_labels is not None:
            self.known_labels.add(lbl.upper() if self.caseInsensitive else lbl)
        return lbl



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import os
import re
import sys

from pyomo.common.dependencies import attempt_import
from pyomo.opt import SolverFactory, SolverManagerFactory, OptSolver
from pyomo.opt.parallel.manager import ActionManagerError, ActionStatus
from pyomo.opt.parallel.async_solver import AsynchronousSolverManager
from pyomo.core.base import Block
import pyomo.neos.kestrel

xmlrpc_client = attempt_import('xmlrpc.client')[0]

logger = logging.getLogger('pyomo.neos')


def _neos_error(msg, results, current_message):
    error_re = re.compile('error', flags=re.I)
    warn_re = re.compile('warn', flags=re.I)

    logger.error("%s  NEOS log:\n%s" % (msg, current_message), exc_info=sys.exc_info())
    soln_data = results.data.decode('utf-8')
    for line in soln_data.splitlines():
        if error_re.search(line):
            logger.error(line)
        elif warn_re.search(line):
            logger.warning(line)


@SolverManagerFactory.register(
    'neos', doc="Asynchronously execute solvers on the NEOS server"
)
class SolverManager_NEOS(AsynchronousSolverManager):
    def clear(self):
        """
        Clear manager state
        """
        AsynchronousSolverManager.clear(self)
        self.kestrel = pyomo.neos.kestrel.kestrelAMPL()
        self._ah = {}  # maps NEOS job numbers to their corresponding
        # action handle.
        self._args = {}
        self._opt_data = {}

        # to grab streamed output from NEOS, need to keep
        # map of action handle to the to-date string of
        # extracted output.
        # TBD: The following entries aren't currently cleaned up, but
        #      we're still trying to get the basics down.
        # store pairs of NEOS message offset and NEOS message string.
        # index into the map is the NEOS job number
        self._neos_log = {}
        self._solvers = {}

    def _perform_queue(self, ah, *args, **kwds):
        """
        Perform the queue operation.  This method returns the ActionHandle,
        and the ActionHandle status indicates whether the queue was successful.
        """
        solver = kwds.pop('solver', kwds.pop('opt', None))
        if solver is None:
            raise ActionManagerError(
                "No solver passed to %s, use keyword option 'solver'"
                % (type(self).__name__)
            )
        if not isinstance(solver, str):
            solver_name = solver.name
            if solver_name == 'asl':
                solver_name = os.path.basename(solver.executable())
        else:
            solver_name = solver
            solver = None

        #
        # Handle ephemeral solvers options here. These
        # will override whatever is currently in the options
        # dictionary, but we will reset these options to
        # their original value at the end of this method.
        #
        user_solver_options = {}
        # make sure to transfer the options dict on the
        # solver plugin if the user does not use a string
        # to identify the neos solver. The ephemeral
        # options must also go after these.
        if solver is not None:
            user_solver_options.update(solver.options)
        _options = kwds.pop('options', {})
        if isinstance(_options, str):
            _options = OptSolver._options_string_to_dict(_options)
        user_solver_options.update(_options)
        user_solver_options.update(
            OptSolver._options_string_to_dict(kwds.pop('options_string', ''))
        )

        # JDS: [5/13/17] The following is a HACK.  This timeout flag is
        # set by pyomo/scripting/util.py:apply_optimizer.  If we do not
        # remove it, it will get passed to the NEOS solver.  For solvers
        # like CPLEX 12.7.0, this will cause a fatal error as it is not
        # a known option.
        if user_solver_options.get('timelimit', 0) is None:
            del user_solver_options['timelimit']

        opt = SolverFactory('_neos')
        opt._presolve(*args, **kwds)
        #
        # Map NEOS name, using lowercase convention in Pyomo
        #
        if len(self._solvers) == 0:
            for name in self.kestrel.solvers():
                if name.endswith('AMPL'):
                    self._solvers[name[:-5].lower()] = name[:-5]
        if solver_name not in self._solvers:
            raise ActionManagerError(
                "Solver '%s' is not recognized by NEOS. "
                "Solver names recognized:\n%s"
                % (solver_name, str(sorted(self._solvers.keys())))
            )
        #
        # Apply kestrel
        #
        # Set the kestrel_options environment
        #
        neos_sname = self._solvers[solver_name].lower()
        os.environ['kestrel_options'] = 'solver=%s' % self._solvers[solver_name]
        #
        # Set the <solver>_options environment
        #
        solver_options = {}
        for key in opt.options:
            solver_options[key] = opt.options[key]
        solver_options.update(user_solver_options)
        options = opt._get_options_string(solver_options)
        if not options == "":
            os.environ[neos_sname + '_options'] = options
        #
        # Generate an XML string using these two environment variables
        #
        xml = self.kestrel.formXML(opt._problem_files[0])
        jobNumber, password = self.kestrel.submit(xml)
        ah.job = jobNumber
        ah.password = password
        #
        # Cleanup
        #
        del os.environ['kestrel_options']
        try:
            del os.environ[neos_sname + "_options"]
        except:
            pass
        #
        # Store action handle, and return
        #
        self._ah[jobNumber] = ah
        self._neos_log[jobNumber] = (0, "")
        self._opt_data[jobNumber] = (
            opt,
            opt._smap_id,
            opt._load_solutions,
            opt._select_index,
            opt._default_variable_value,
        )
        self._args[jobNumber] = args
        return ah

    def _perform_wait_any(self):
        """
        Perform the wait_any operation.  This method returns an
        ActionHandle with the results of waiting.  If None is returned
        then the ActionManager assumes that it can call this method again.
        Note that an ActionHandle can be returned with a dummy value,
        to indicate an error.
        """
        for jobNumber in self._ah:
            status = self.kestrel.neos.getJobStatus(
                jobNumber, self._ah[jobNumber].password
            )

            if status not in ("Running", "Waiting"):
                # the job is done.
                ah = self._ah[jobNumber]
                del self._ah[jobNumber]
                ah.status = ActionStatus.done

                opt, smap_id, load_solutions, select_index, default_variable_value = (
                    self._opt_data[jobNumber]
                )
                del self._opt_data[jobNumber]

                args = self._args[jobNumber]
                del self._args[jobNumber]

                # retrieve the final results, which are in message/log format.
                results = self.kestrel.neos.getFinalResults(jobNumber, ah.password)

                current_offset, current_message = self._neos_log[jobNumber]
                with open(opt._log_file, 'w') as OUTPUT:
                    OUTPUT.write(current_message)
                with open(opt._soln_file, 'w') as OUTPUT:
                    OUTPUT.write(results.data.decode('utf-8'))

                rc = None
                try:
                    solver_results = opt.process_output(rc)
                except:
                    _neos_error(
                        "Error parsing NEOS solution file", results, current_message
                    )
                    return ah

                solver_results._smap_id = smap_id
                self.results[ah.id] = solver_results

                if isinstance(args[0], Block):
                    _model = args[0]
                    if load_solutions:
                        try:
                            _model.solutions.load_from(
                                solver_results,
                                select=select_index,
                                default_variable_value=default_variable_value,
                            )
                        except:
                            _neos_error(
                                "Error loading NEOS solution into model",
                                results,
                                current_message,
                            )
                        solver_results._smap_id = None
                        solver_results.solution.clear()
                    else:
                        solver_results._smap = _model.solutions.symbol_map[smap_id]
                        _model.solutions.delete_symbol_map(smap_id)

                return ah
            else:
                # The job is still running...
                #
                # Grab the partial messages from NEOS as you go, in case
                # you want to output on-the-fly. You will only get data
                # if the job was routed to the "short" priority queue.
                current_offset, current_message = self._neos_log[jobNumber]
                # TBD: blocking isn't the way to go, but non-blocking
                # was triggering some exception in kestrel.
                #
                # [5/13/17]: The blocking fetch will timeout in 2
                # minutes.  If NEOS doesn't produce intermediate results
                # by then we will need to catch (and eat) the exception
                try:
                    message_fragment, new_offset = (
                        self.kestrel.neos.getIntermediateResults(
                            jobNumber, self._ah[jobNumber].password, current_offset
                        )
                    )
                    logger.info(message_fragment)
                    self._neos_log[jobNumber] = (
                        new_offset,
                        current_message + ((message_fragment.data).decode('utf-8')),
                    )
                except xmlrpc_client.ProtocolError:
                    # The command probably timed out
                    pass

        return None

    def _kill_all_pending_jobs(self):
        for ah in self._ah.values():
            self.kestrel.kill(ah.job, ah.password)

    def __exit__(self, t, v, traceback):
        self._kill_all_pending_jobs()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.common.dependencies import numpy, numpy_available
from pyomo.common.log import LoggingIntercept

# Boolean
numpy_bool_names = []
if numpy_available:
    if numpy.__version__[0] == '2':
        numpy_bool_names.append('bool')
    else:
        numpy_bool_names.append('bool_')
# Integers
numpy_int_names = []
if numpy_available:
    numpy_int_names.append('int_')
    numpy_int_names.append('intc')
    numpy_int_names.append('intp')
    numpy_int_names.append('int8')
    numpy_int_names.append('int16')
    numpy_int_names.append('int32')
    numpy_int_names.append('int64')
    numpy_int_names.append('uint8')
    numpy_int_names.append('uint16')
    numpy_int_names.append('uint32')
    numpy_int_names.append('uint64')
# Reals
numpy_float_names = []
if numpy_available:
    if hasattr(numpy, 'float_'):
        numpy_float_names.append('float_')
    numpy_float_names.append('float16')
    numpy_float_names.append('float32')
    numpy_float_names.append('float64')
    if hasattr(numpy, 'float96'):
        numpy_float_names.append('float96')
    if hasattr(numpy, 'float128'):
        # On some numpy builds, the name of float128 is longdouble
        numpy_float_names.append(numpy.float128.__name__)
# Complex
numpy_complex_names = []
if numpy_available:
    if hasattr(numpy, 'complex_'):
        numpy_complex_names.append('complex_')
    numpy_complex_names.append('complex64')
    numpy_complex_names.append('complex128')
    if hasattr(numpy, 'complex192'):
        numpy_complex_names.append('complex192')
    if hasattr(numpy, 'complex256'):
        # On some numpy builds, the name of complex256 is clongdouble
        numpy_complex_names.append(numpy.complex256.__name__)


class TestNumpyRegistration(unittest.TestCase):
    def test_deprecation(self):
        with LoggingIntercept() as LOG:
            import pyomo.core.kernel.register_numpy_types as rnt
        self.assertRegex(
            LOG.getvalue(),
            "DEPRECATED: pyomo.core.kernel.register_numpy_types is deprecated.",
        )
        self.assertEqual(sorted(rnt.numpy_bool_names), sorted(numpy_bool_names))
        self.assertEqual(sorted(rnt.numpy_int_names), sorted(numpy_int_names))
        self.assertEqual(sorted(rnt.numpy_float_names), sorted(numpy_float_names))
        self.assertEqual(sorted(rnt.numpy_complex_names), sorted(numpy_complex_names))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Development of this module was conducted as part of the Institute for
# the Design of Advanced Energy Systems (IDAES) with support through the
# Simulation-Based Engineering, Crosscutting Research Program within the
# U.S. Department of Energy's Office of Fossil Energy and Carbon Management.
# ____________________________________________________________________________________

import logging

import pyomo.common.unittest as unittest
from pyomo.common.collections import ComponentMap, ComponentSet
from pyomo.environ import (
    Var,
    VarList,
    ConcreteModel,
    Reals,
    ExternalFunction,
    value,
    Objective,
    Constraint,
    sqrt,
    sin,
    cos,
    SolverFactory,
)
from pyomo.core.base.var import VarData
from pyomo.core.expr.numeric_expr import ExternalFunctionExpression
from pyomo.core.expr.visitor import identify_variables
from pyomo.contrib.trustregion.interface import TRFInterface
from pyomo.contrib.trustregion.TRF import _trf_config

logger = logging.getLogger('pyomo.contrib.trustregion')


class TestTrustRegionInterface(unittest.TestCase):
    def setUp(self):
        self.m = ConcreteModel()
        self.m.z = Var(range(3), domain=Reals, initialize=2.0)
        self.m.x = Var(range(2), initialize=2.0)
        self.m.x[1] = 1.0

        def blackbox(a, b):
            return sin(a - b)

        def grad_blackbox(args, fixed):
            a, b = args[:2]
            return [cos(a - b), -cos(a - b)]

        self.m.bb = ExternalFunction(blackbox, grad_blackbox)

        self.m.obj = Objective(
            expr=(self.m.z[0] - 1.0) ** 2
            + (self.m.z[0] - self.m.z[1]) ** 2
            + (self.m.z[2] - 1.0) ** 2
            + (self.m.x[0] - 1.0) ** 4
            + (self.m.x[1] - 1.0) ** 6
        )
        self.m.c1 = Constraint(
            expr=(
                self.m.x[0] * self.m.z[0] ** 2 + self.m.bb(self.m.x[0], self.m.x[1])
                == 2 * sqrt(2.0)
            )
        )
        self.m.c2 = Constraint(
            expr=self.m.z[2] ** 4 * self.m.z[1] ** 2 + self.m.z[1] == 8 + sqrt(2.0)
        )
        self.config = _trf_config()
        self.ext_fcn_surrogate_map_rule = lambda comp, ef: 0
        self.interface = TRFInterface(
            self.m,
            [self.m.z[0], self.m.z[1], self.m.z[2]],
            self.ext_fcn_surrogate_map_rule,
            self.config,
        )

    def test_initializeInterface(self):
        self.assertEqual(self.m, self.interface.original_model)
        self.assertEqual(self.config, self.interface.config)
        self.assertEqual(
            self.interface.basis_expression_rule, self.ext_fcn_surrogate_map_rule
        )
        self.assertEqual('ipopt', self.interface.solver.name)

    def test_replaceRF(self):
        # These data objects are normally initialized by
        # replaceExternalFunctionsWithVariables
        self.interface.data.all_variables = ComponentSet()
        self.interface.data.truth_models = ComponentMap()
        self.interface.data.ef_outputs = VarList()
        # The objective function has no EF.
        # Therefore, replaceEF should do nothing
        expr = self.interface.model.obj.expr
        new_expr = self.interface.replaceEF(expr)
        self.assertEqual(expr, new_expr)
        # The first constraint has one EF.
        # Therefore, replaceEF should do a substitution
        expr = self.interface.model.c1.expr
        new_expr = self.interface.replaceEF(expr)
        self.assertIsNot(expr, new_expr)
        self.assertEqual(
            str(new_expr),
            'x[0]*z[0]**2 + trf_data.ef_outputs[1]  ==  2.8284271247461903',
        )

    def test_remove_ef_from_expr(self):
        # These data objects are normally initialized by
        # replaceExternalFunctionsWithVariables
        self.interface.data.all_variables = ComponentSet()
        self.interface.data.truth_models = ComponentMap()
        self.interface.data.ef_outputs = VarList()
        self.interface.data.basis_expressions = ComponentMap()
        # The objective function has no EF.
        # Therefore, remove_ef_from_expr should do nothing
        component = self.interface.model.obj
        self.interface._remove_ef_from_expr(component)
        self.assertEqual(
            str(self.interface.model.obj.expr),
            '(z[0] - 1.0)**2 + (z[0] - z[1])**2 + (z[2] - 1.0)**2 + (x[0] - 1.0)**4 + (x[1] - 1.0)**6',
        )
        # The first constraint has one EF.
        # Therefore, remove_ef_from_expr should do something
        component = self.interface.model.c1
        str_expr = str(component.expr)
        self.interface._remove_ef_from_expr(component)
        self.assertNotEqual(str_expr, str(component.expr))
        self.assertEqual(
            str(component.expr),
            'x[0]*z[0]**2 + trf_data.ef_outputs[1]  ==  2.8284271247461903',
        )

    def test_replaceExternalFunctionsWithVariables(self):
        # In running this method, we not only replace EFs
        # with 'holder' vars; we also get useful information
        # about inputs, outputs, basis expressions, etc.
        self.interface.replaceExternalFunctionsWithVariables()
        # Check the directly defined model vars against all_variables
        for var in self.interface.model.component_data_objects(Var):
            self.assertIn(var, ComponentSet(self.interface.data.all_variables))
        # Check the output vars against all_variables
        for i in self.interface.data.ef_outputs:
            self.assertIn(
                self.interface.data.ef_outputs[i],
                ComponentSet(self.interface.data.all_variables),
            )
        # The truth models should be a mapping from the EF to
        # the replacement
        for i, k in self.interface.data.truth_models.items():
            self.assertIsInstance(k, ExternalFunctionExpression)
            self.assertIn(str(self.interface.model.x[0]), str(k))
            self.assertIn(str(self.interface.model.x[1]), str(k))
            self.assertIsInstance(i, VarData)
            self.assertEqual(i, self.interface.data.ef_outputs[1])
        for i, k in self.interface.data.basis_expressions.items():
            self.assertEqual(k, 0)
            self.assertEqual(i, self.interface.data.ef_outputs[1])
        self.assertEqual(1, list(self.interface.data.ef_inputs.keys())[0])
        self.assertEqual(
            self.interface.data.ef_inputs[1],
            [self.interface.model.x[0], self.interface.model.x[1]],
        )
        # HACK: This was in response to a hack.
        # Remove when NL writer re-write is complete.
        # Make sure that EFs were removed from the cloned model.
        self.assertEqual(
            list(self.interface.model.component_objects(ExternalFunction)), []
        )
        # TRF only supports one active Objective.
        # Make sure that it fails if there are multiple objs.
        self.m.obj2 = Objective(expr=(self.m.x[0] ** 2 - (self.m.z[1] - 3) ** 3))
        interface = TRFInterface(
            self.m,
            [self.m.z[0], self.m.z[1], self.m.z[2]],
            self.ext_fcn_surrogate_map_rule,
            self.config,
        )
        with self.assertRaises(ValueError):
            interface.replaceExternalFunctionsWithVariables()

    def test_createConstraints(self):
        # replaceExternalFunctionsWithVariables sets up some
        # necessary items in the block
        self.interface.replaceExternalFunctionsWithVariables()
        self.interface.createConstraints()
        # The constraints should have been created and deactivated
        self.assertFalse(self.interface.data.basis_constraint.active)
        self.assertFalse(self.interface.data.sm_constraint_basis.active)
        # The size of each constraint should be 1
        self.assertEqual(len(self.interface.data.basis_constraint), 1)
        self.assertEqual(len(self.interface.data.sm_constraint_basis), 1)
        # Because they are size 1, they should have one key
        self.assertEqual(list(self.interface.data.basis_constraint.keys()), [1])
        cs = ComponentSet(
            identify_variables(self.interface.data.basis_constraint[1].expr)
        )
        # The basis constraint only has the EF variable
        self.assertEqual(len(cs), 1)
        self.assertIn(self.interface.data.ef_outputs[1], cs)
        cs = ComponentSet(
            identify_variables(self.interface.data.sm_constraint_basis[1].expr)
        )
        # The surrogate model constraint has the EF var, with inputs
        # of x[0] and x[1], as seen in self.m.c1
        self.assertEqual(len(cs), 3)
        self.assertIn(self.interface.model.x[0], cs)
        self.assertIn(self.interface.model.x[1], cs)
        self.assertIn(self.interface.data.ef_outputs[1], cs)

    def test_updateSurrogateModel(self):
        # Set up necessary data objects and Params
        self.interface.replaceExternalFunctionsWithVariables()
        self.interface.createConstraints()
        # Set starter values on the model
        self.interface.model.x[0] = 2.0
        self.interface.model.z.set_values({0: 5.0, 1: 2.5, 2: -1.0})
        self.interface.data.basis_model_output[:] = 0
        self.interface.data.grad_basis_model_output[...] = 0
        self.interface.data.truth_model_output[:] = 0
        self.interface.data.grad_truth_model_output[...] = 0
        self.interface.data.value_of_ef_inputs[...] = 0
        self.interface.updateSurrogateModel()
        # The basis model here is 0, so all basis values should be 0
        for key, val in self.interface.data.basis_model_output.items():
            self.assertEqual(value(val), 0)
        for key, val in self.interface.data.grad_basis_model_output.items():
            self.assertEqual(value(val), 0)
        for key, val in self.interface.data.truth_model_output.items():
            self.assertAlmostEqual(value(val), 0.8414709848078965)
        # The truth gradients should equal the output of [cos(2-1), -cos(2-1)]
        truth_grads = []
        for key, val in self.interface.data.grad_truth_model_output.items():
            truth_grads.append(value(val))
        self.assertEqual(truth_grads, [cos(1), -cos(1)])
        # The inputs should be the values of x[0] and x[1]
        for key, val in self.interface.data.value_of_ef_inputs.items():
            self.assertEqual(value(self.interface.model.x[key[1]]), value(val))
        # Change the model values to something else and try again
        self.interface.model.x.set_values({0: 0, 1: 0})
        self.interface.updateSurrogateModel()
        # The basis values should still all be 0
        for key, val in self.interface.data.basis_model_output.items():
            self.assertEqual(value(val), 0)
        for key, val in self.interface.data.grad_basis_model_output.items():
            self.assertEqual(value(val), 0)
        # We still have not updated this value, so the value should be 0
        for key, val in self.interface.data.truth_model_output.items():
            self.assertEqual(value(val), 0)
        # The truth gradients should equal the output of [cos(0-0), -cos(0-0)]
        truth_grads = []
        for key, val in self.interface.data.grad_truth_model_output.items():
            truth_grads.append(value(val))
        self.assertEqual(truth_grads, [cos(0), -cos(0)])
        # The inputs should be the values of x[0] and x[1]
        for key, val in self.interface.data.value_of_ef_inputs.items():
            self.assertEqual(value(self.interface.model.x[key[1]]), value(val))

    def test_getCurrentDecisionVariableValues(self):
        # Set up necessary data objects
        self.interface.replaceExternalFunctionsWithVariables()
        # Set starter values on the model
        self.interface.model.x[0] = 2.0
        self.interface.model.z.set_values({0: 5.0, 1: 2.5, 2: -1.0})
        current_values = self.interface.getCurrentDecisionVariableValues()
        for var in self.interface.decision_variables:
            self.assertIn(var.name, list(current_values.keys()))
            self.assertEqual(current_values[var.name], value(var))

    @unittest.skipIf(
        not SolverFactory('ipopt').available(False), "The IPOPT solver is not available"
    )
    def test_updateDecisionVariableBounds(self):
        # Initialize the problem
        self.interface.initializeProblem()
        # Make sure the initial bounds match the current bounds
        for var in self.interface.decision_variables:
            self.assertEqual(
                self.interface.initial_decision_bounds[var.name], [var.lb, var.ub]
            )
        # Update the bounds and make sure that the initial no longer match
        # the current bounds
        self.interface.updateDecisionVariableBounds(0.5)
        for var in self.interface.decision_variables:
            self.assertNotEqual(
                self.interface.initial_decision_bounds[var.name], [var.lb, var.ub]
            )

    def test_getCurrentModelState(self):
        # Set up necessary data objects
        self.interface.replaceExternalFunctionsWithVariables()
        # Set starter values on the model
        self.interface.model.x[0] = 2.0
        self.interface.model.z.set_values({0: 5.0, 1: 2.5, 2: -1.0})
        result = self.interface.getCurrentModelState()
        self.assertEqual(len(result), len(self.interface.data.all_variables))
        for var in self.interface.data.all_variables:
            self.assertIn(value(var), result)

    @unittest.skipIf(
        not SolverFactory('ipopt').available(False), "The IPOPT solver is not available"
    )
    def test_calculateFeasibility(self):
        # Set up necessary data objects
        self.interface.replaceExternalFunctionsWithVariables()
        self.interface.createConstraints()
        # Set starter values on the model
        self.interface.model.x[0] = 2.0
        self.interface.model.z.set_values({0: 5.0, 1: 2.5, 2: -1.0})
        self.interface.data.basis_model_output[:] = 0
        self.interface.data.grad_basis_model_output[...] = 0
        self.interface.data.truth_model_output[:] = 0
        self.interface.data.grad_truth_model_output[...] = 0
        self.interface.data.value_of_ef_inputs[...] = 0
        feasibility = self.interface.calculateFeasibility()
        # The initial feasibility should be 0 because we haven't
        # solved anything, so the truth model and ef_outputs are the same
        self.assertEqual(feasibility, 0)
        # We update the surrogate model to get real parameters
        self.interface.updateSurrogateModel()
        feasibility = self.interface.calculateFeasibility()
        # Because we have not solved the model, the output and truth model
        # should currently match
        self.assertEqual(feasibility, 0)
        # Check after a solve is completed
        self.interface.data.basis_constraint.activate()
        objective, step_norm, feasibility = self.interface.solveModel()
        self.assertAlmostEqual(feasibility, 0.09569982275514467)
        self.interface.data.basis_constraint.deactivate()

    @unittest.skipIf(
        not SolverFactory('ipopt').available(False), "The IPOPT solver is not available"
    )
    def test_calculateStepSizeInfNorm(self):
        # Set up necessary data objects
        self.interface.replaceExternalFunctionsWithVariables()
        self.interface.createConstraints()
        # Set starter values on the model
        self.interface.model.x[0] = 2.0
        self.interface.model.z.set_values({0: 5.0, 1: 2.5, 2: -1.0})
        self.interface.data.basis_model_output[:] = 0
        self.interface.data.grad_basis_model_output[...] = 0
        self.interface.data.truth_model_output[:] = 0
        self.interface.data.grad_truth_model_output[...] = 0
        self.interface.data.value_of_ef_inputs[...] = 0
        # Get original and new values
        original_values = self.interface.getCurrentDecisionVariableValues()
        self.interface.updateSurrogateModel()
        new_values = self.interface.getCurrentDecisionVariableValues()
        stepnorm = self.interface.calculateStepSizeInfNorm(original_values, new_values)
        # Currently, we have taken NO step.
        # Therefore, the norm should be 0.
        self.assertEqual(stepnorm, 0)
        # Check after a solve is completed
        self.interface.data.basis_constraint.activate()
        objective, step_norm, feasibility = self.interface.solveModel()
        self.assertAlmostEqual(step_norm, 3.393437471478297)
        self.interface.data.basis_constraint.deactivate()

    @unittest.skipIf(
        not SolverFactory('ipopt').available(False), "The IPOPT solver is not available"
    )
    def test_solveModel(self):
        # Set up initial data objects and Params
        self.interface.replaceExternalFunctionsWithVariables()
        self.interface.createConstraints()
        self.interface.data.basis_constraint.activate()
        # Set starter values on the model
        self.interface.model.x[0] = 2.0
        self.interface.model.z.set_values({0: 5.0, 1: 2.5, 2: -1.0})
        self.interface.data.basis_model_output[:] = 0
        self.interface.data.grad_basis_model_output[...] = 0
        self.interface.data.truth_model_output[:] = 0
        self.interface.data.grad_truth_model_output[...] = 0
        self.interface.data.value_of_ef_inputs[...] = 0
        # Run the solve
        objective, step_norm, feasibility = self.interface.solveModel()
        self.assertAlmostEqual(objective, 5.150744273013601)
        self.assertAlmostEqual(step_norm, 3.393437471478297)
        self.assertAlmostEqual(feasibility, 0.09569982275514467)
        self.interface.data.basis_constraint.deactivate()
        # Change the constraint and update the surrogate model
        self.interface.updateSurrogateModel()
        self.interface.data.sm_constraint_basis.activate()
        objective, step_norm, feasibility = self.interface.solveModel()
        self.assertAlmostEqual(objective, 5.15065981284333)
        self.assertAlmostEqual(step_norm, 0.0017225116628372117)
        self.assertAlmostEqual(feasibility, 0.00014665023773349772)

    @unittest.skipIf(
        not SolverFactory('ipopt').available(False), "The IPOPT solver is not available"
    )
    def test_initializeProblem(self):
        # Set starter values on the model
        self.interface.model.x[0] = 2.0
        self.interface.model.z.set_values({0: 5.0, 1: 2.5, 2: -1.0})
        objective, feasibility = self.interface.initializeProblem()
        for var in self.interface.decision_variables:
            self.assertIn(var.name, list(self.interface.initial_decision_bounds.keys()))
            self.assertEqual(
                self.interface.initial_decision_bounds[var.name], [var.lb, var.ub]
            )
        self.assertAlmostEqual(objective, 5.150744273013601)
        self.assertAlmostEqual(feasibility, 0.09569982275514467)
        self.assertTrue(self.interface.data.sm_constraint_basis.active)
        self.assertFalse(self.interface.data.basis_constraint.active)

    @unittest.skipIf(
        not SolverFactory('ipopt').available(False), "The IPOPT solver is not available"
    )
    def test_rejectStep(self):
        self.interface.model.x[1] = 1.5
        self.interface.model.x[0] = 2.0
        self.interface.model.z.set_values({0: 5.0, 1: 2.5, 2: -1.0})
        self.interface.replaceExternalFunctionsWithVariables()
        self.interface.createConstraints()
        self.interface.data.basis_constraint.activate()
        _, _, _ = self.interface.solveModel()
        self.assertEqual(
            len(self.interface.data.all_variables),
            len(self.interface.data.previous_model_state),
        )
        # Make sure the values changed from the original model
        self.assertNotEqual(value(self.interface.model.x[0]), 2.0)
        self.assertNotEqual(value(self.interface.model.x[1]), 1.5)
        self.assertNotEqual(value(self.interface.model.z[0]), 5.0)
        self.assertNotEqual(value(self.interface.model.z[1]), 2.5)
        self.assertNotEqual(value(self.interface.model.z[2]), -1.0)
        self.interface.rejectStep()
        # Make sure the values were reset
        self.assertEqual(value(self.interface.model.x[0]), 2.0)
        self.assertEqual(value(self.interface.model.x[1]), 1.5)
        self.assertEqual(value(self.interface.model.z[0]), 5.0)
        self.assertEqual(value(self.interface.model.z[1]), 2.5)
        self.assertEqual(value(self.interface.model.z[2]), -1.0)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# This module was originally developed as part of the IDAES PSE Framework
#
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
A simple GUI viewer/editor for Pyomo models.
"""

__author__ = "John Eslick"

import os
import logging

from pyomo.common.fileutils import this_file_dir
from pyomo.common.flags import building_documentation
from pyomo.contrib.viewer.report import value_no_exception, get_residual
from pyomo.core.base.param import ParamData
from pyomo.environ import (
    Block,
    BooleanVar,
    Var,
    Constraint,
    Param,
    Expression,
    value,
    units,
)

import pyomo.contrib.viewer.qt as myqt

_log = logging.getLogger(__name__)


# This lets the file be imported when the Qt UI is not available (or
# when building docs), but you won't be able to use it
class _ModelBrowserUI:
    pass


class _ModelBrowser:
    pass


# Note that the classes loaded here have signatures that are not
# parsable by Sphinx, so we won't attempt to import them if we are
# building the API documentation.
if not building_documentation():
    import sys

    mypath = this_file_dir()
    try:
        _ModelBrowserUI, _ModelBrowser = myqt.uic.loadUiType(
            os.path.join(mypath, "model_browser.ui")
        )
    except:
        pass


class LineEditCreator(myqt.QItemEditorCreatorBase):
    """
    Class to create editor widget for int and floats in a model view type object
    """

    def createWidget(self, parent):
        return myqt.QLineEdit(parent=parent)


class NumberDelegate(myqt.QItemDelegate):
    """
    Tree view item delegate. This is used here to change how items are edited.
    """

    def __init__(self, parent):
        super().__init__(parent=parent)
        factory = myqt.QItemEditorFactory()
        factory.registerEditor(myqt.QMetaType.Int, LineEditCreator())
        factory.registerEditor(myqt.QMetaType.Double, LineEditCreator())
        self.setItemEditorFactory(factory)

    def setModelData(self, editor, model, index):
        if isinstance(editor, myqt.QComboBox):
            value = editor.currentText()
        else:
            value = editor.text()
        a = model.column[index.column()]
        isinstance(index.internalPointer().get(a), bool)
        try:  # Recognize ints and floats.
            if value == "False" or value == "false":
                index.internalPointer().set(a, False)
            elif value == "True" or value == "true":
                index.internalPointer().set(a, True)
            elif "." in value or "e" in value or "E" in value:
                index.internalPointer().set(a, float(value))
            else:
                index.internalPointer().set(a, int(value))
        except:  # If not a valid number ignore
            pass


class ModelBrowser(_ModelBrowser, _ModelBrowserUI):
    def __init__(self, ui_data, standard="Var"):
        """
        Create a dock widget with a QTreeView of a Pyomo model.

        Args:
            ui_data: Contains model and ui information
            standard: A standard setup for different types of model components
                {"Var", "Constraint", "Param", "Expression"}
        """
        super().__init__()
        self.setupUi(self)
        # The default int and double spin boxes are not good for this
        # application.  So just use regular line edits.
        number_delegate = NumberDelegate(self)
        self.ui_data = ui_data
        self.ui_data.updated.connect(self.update_model)
        self.treeView.setItemDelegate(number_delegate)
        if standard == "Var":
            # This if block sets up standard views
            components = (Var, BooleanVar)
            columns = ["name", "value", "ub", "lb", "fixed", "stale", "units", "domain"]
            editable = ["value", "ub", "lb", "fixed"]
            self.setWindowTitle("Variables")
        elif standard == "Constraint":
            components = Constraint
            columns = ["name", "value", "ub", "lb", "residual", "active"]
            editable = ["active"]
            self.setWindowTitle("Constraints")
        elif standard == "Param":
            components = Param
            columns = ["name", "value", "mutable", "units"]
            editable = ["value"]
            self.setWindowTitle("Parameters")
        elif standard == "Expression":
            components = Expression
            columns = ["name", "value", "units"]
            editable = []
            self.setWindowTitle("Expressions")
        else:
            raise ValueError("{} is not a valid view type".format(standard))
        # Create a data model.  This is what translates the Pyomo model into
        # a tree view.
        datmodel = ComponentDataModel(
            self,
            ui_data=ui_data,
            columns=columns,
            components=components,
            editable=editable,
        )
        self.datmodel = datmodel
        self.treeView.setModel(datmodel)
        self.treeView.setColumnWidth(0, 400)
        # Selection behavior: select a whole row, can select multiple rows.
        self.treeView.setSelectionBehavior(
            myqt.QAbstractItemView.SelectionBehavior.SelectRows
        )
        self.treeView.setSelectionMode(
            myqt.QAbstractItemView.SelectionMode.ExtendedSelection
        )

    def refresh(self):
        added = self.datmodel._update_tree()
        self.datmodel.layoutChanged.emit()

    def update_model(self):
        self.datmodel.update_model()


class ComponentDataItem:
    """
    This is a container for a Pyomo component to be displayed in a model tree
    view.

    Args:
        parent: parent data item
        o: pyomo component object
        ui_data: a container for data, as of now mainly just pyomo model
    """

    def __init__(self, parent, o, ui_data):
        self.ui_data = ui_data
        self.data = o
        self.parent = parent
        self.children = []  # child items
        self.ids = {}
        self.get_callback = {
            "value": self._get_value_callback,
            "lb": self._get_lb_callback,
            "ub": self._get_ub_callback,
            "expr": self._get_expr_callback,
            "residual": self._get_residual_callback,
            "units": self._get_units_callback,
            "domain": self._get_domain_callback,
        }
        self.set_callback = {
            "value": self._set_value_callback,
            "lb": self._set_lb_callback,
            "ub": self._set_ub_callback,
            "active": self._set_active_callback,
            "fixed": self._set_fixed_callback,
        }

    @property
    def _cache_value(self):
        return self.ui_data.value_cache.get(self.data, None)

    @property
    def _cache_units(self):
        return self.ui_data.value_cache_units.get(self.data, None)

    def add_child(self, o):
        """Add a child data item"""
        item = ComponentDataItem(self, o, ui_data=self.ui_data)
        self.children.append(item)
        self.ids[id(o)] = item
        return item

    def get(self, a):
        """Get an attribute"""
        if a in self.get_callback:
            return self.get_callback[a]()
        else:
            try:
                return getattr(self.data, a)
            except:
                return None

    def set(self, a, val):
        """set an attribute"""
        if a in self.set_callback:
            return self.set_callback[a](val)
        else:
            try:
                return setattr(self.data, a, val)
            except:
                _log.exception("Can't set value of {}".format(a))
                return None

    def _get_expr_callback(self):
        if hasattr(self.data, "expr"):
            return str(self.data.expr)
        else:
            return None

    def _get_value_callback(self):
        if isinstance(self.data, ParamData):
            v = value_no_exception(self.data, div0="divide_by_0")
            # Check the param value for numpy float and int, sometimes numpy
            # values can sneak in especially if you set parameters from data
            # and for whatever reason numpy values don't display
            if isinstance(v, float):  # includes numpy float
                v = float(v)
            elif isinstance(v, int):  # includes numpy int
                v = int(v)
            return v
        elif isinstance(
            self.data, (Var._ComponentDataClass, BooleanVar._ComponentDataClass)
        ):
            v = value_no_exception(self.data)
            # Check the param value for numpy float and int, sometimes numpy
            # values can sneak in especially if you set parameters from data
            # and for whatever reason numpy values don't display
            if isinstance(v, float):  # includes numpy float
                v = float(v)
            elif isinstance(v, int):  # includes numpy int
                v = int(v)
            return v
        elif isinstance(self.data, (float, int)):
            return self.data
        else:
            return self._cache_value

    def _get_lb_callback(self):
        if isinstance(self.data, (Var._ComponentDataClass)):
            return self.data.lb
        elif hasattr(self.data, "lower"):
            return value_no_exception(self.data.lower, div0="Divide_by_0")
        else:
            return None

    def _get_ub_callback(self):
        if isinstance(self.data, (Var._ComponentDataClass)):
            return self.data.ub
        elif hasattr(self.data, "upper"):
            return value_no_exception(self.data.upper, div0="Divide_by_0")
        else:
            return None

    def _get_residual_callback(self):
        if isinstance(self.data, Constraint._ComponentDataClass):
            return get_residual(self.ui_data, self.data)
        else:
            return None

    def _get_units_callback(self):
        if isinstance(self.data, (Var, Var._ComponentDataClass)):
            return str(units.get_units(self.data))
        if isinstance(self.data, (Param, ParamData)):
            return str(units.get_units(self.data))
        return self._cache_units

    def _get_domain_callback(self):
        if isinstance(self.data, Var._ComponentDataClass):
            return str(self.data.domain)
        if isinstance(self.data, (BooleanVar, BooleanVar._ComponentDataClass)):
            return "BooleanVar"
        return None

    def _set_value_callback(self, val):
        if isinstance(
            self.data, (Var._ComponentDataClass, BooleanVar._ComponentDataClass)
        ):
            try:
                self.data.value = val
            except:
                return
        elif isinstance(self.data, (Var, BooleanVar)):
            try:
                for o in self.data.values():
                    o.value = val
            except:
                return
        elif isinstance(self.data, ParamData):
            if not self.data.parent_component().mutable:
                return
            try:
                self.data.value = val
            except:
                return
        elif isinstance(self.data, Param):
            if not self.data.parent_component().mutable:
                return
            try:
                for o in self.data.values():
                    o.value = val
            except:
                return

    def _set_lb_callback(self, val):
        if isinstance(self.data, (Var._ComponentDataClass)):
            try:
                self.data.setlb(val)
            except:
                return
        elif isinstance(self.data, Var):
            try:
                for o in self.data.values():
                    o.setlb(val)
            except:
                return

    def _set_ub_callback(self, val):
        if isinstance(self.data, (Var._ComponentDataClass)):
            try:
                self.data.setub(val)
            except:
                return
        elif isinstance(self.data, Var):
            try:
                for o in self.data.values():
                    o.setub(val)
            except:
                return

    def _set_active_callback(self, val):
        if not val or val in ["False", "false", "0", "f", "F"]:
            # depending on the version of Qt, you may see a combo box that
            # lets you select true/false or may be able to type the combo
            # box will return True or False, if you have to type could be
            # something else
            val = False
        else:
            val = True
        try:
            if val:
                self.data.activate()
            else:
                self.data.deactivate()
        except:
            return

    def _set_fixed_callback(self, val):
        if not val or val in ["False", "false", "0", "f", "F"]:
            # depending on the version of Qt, you may see a combo box that
            # lets you select true/false or may be able to type the combo
            # box will return True or False, if you have to type could be
            # something else
            val = False
        else:
            val = True
        try:
            if val:
                self.data.fix()
            else:
                self.data.unfix()
        except:
            return


class ComponentDataModel(myqt.QAbstractItemModel):
    """
    This is a data model to provide the tree structure and information
    to the tree viewer
    """

    def __init__(
        self,
        parent,
        ui_data,
        columns=["name", "value"],
        components=(Var, BooleanVar),
        editable=[],
    ):
        super().__init__(parent)
        self.column = columns
        self._col_editable = editable
        self.ui_data = ui_data
        self.components = components
        self.update_model()

    def update_model(self):
        self.rootItems = []
        self._create_tree(o=self.ui_data.model)

    def _update_tree(self, parent=None, o=None):
        """
        Check tree structure against the Pyomo model to add or delete
        components as needed. The arguments are to be used in the recursive
        function. Entering into this don't specify any args.
        """
        # Blocks are special they define the hierarchy of the model, so first
        # check for blocks. Other component can be handled together
        if o is None and len(self.rootItems) > 0:  # top level object (no parent)
            parent = self.rootItems[0]  # should be single root node for now
            o = parent.data  # start with root node
            for no in o.component_objects(descend_into=False):
                # This will traverse the whole Pyomo model tree
                self._update_tree(parent=parent, o=no)
            return
        elif o is None:  # if o is None, but no root nodes (when no model)
            return

        # past the root node go down here
        item = parent.ids.get(id(o), None)
        if item is not None:  # check if any children of item where deleted
            for i in item.children:
                try:
                    if i.data.parent_block() is None:
                        i.parent.children.remove(i)
                        del i.parent.ids[id(i.data)]
                        del i  # probably should descend down and delete stuff
                except AttributeError:
                    # Probably an element of an indexed immutable param
                    pass
        if isinstance(
            o, Block._ComponentDataClass
        ):  # single block or element of indexed block
            if item is None:
                item = self._add_item(parent=parent, o=o)
            for no in o.component_objects(descend_into=False):
                self._update_tree(parent=item, o=no)
        elif isinstance(o, Block):  # indexed block, so need to add elements
            if item is None:
                item = self._add_item(parent=parent, o=o)
            if (
                hasattr(o.index_set(), "is_constructed")
                and o.index_set().is_constructed()
            ):
                for key in sorted(o.keys()):
                    self._update_tree(parent=item, o=o[key])
        elif isinstance(o, self.components):  # anything else
            if item is None:
                item = self._add_item(parent=parent, o=o)
            if (
                hasattr(o.index_set(), "is_constructed")
                and o.index_set().is_constructed()
            ):
                for key in sorted(o.keys()):
                    if key == None:
                        break  # Single variable so skip
                    item2 = item.ids.get(id(o[key]), None)
                    if item2 is None:
                        item2 = self._add_item(parent=item, o=o[key])
                    item2._visited = True
        return

    def _create_tree(self, parent=None, o=None):
        """
        This create a model tree structure to display in a tree view.
        Args:
            parent: a ComponentDataItem underwhich to create a TreeItem
            o: A Pyomo component to add to the tree
        """
        # Blocks are special they define the hierarchy of the model, so first
        # check for blocks. Other component can be handled together
        if isinstance(
            o, Block._ComponentDataClass
        ):  # single block or element of indexed block
            item = self._add_item(parent=parent, o=o)
            for no in o.component_objects(descend_into=False):
                self._create_tree(parent=item, o=no)
        elif isinstance(o, Block):  # indexed block, so need to add elements
            item = self._add_item(parent=parent, o=o)
            if (
                hasattr(o.index_set(), "is_constructed")
                and o.index_set().is_constructed()
            ):
                for key in sorted(o.keys()):
                    self._create_tree(parent=item, o=o[key])
        elif isinstance(o, self.components):  # anything else
            item = self._add_item(parent=parent, o=o)
            if (
                hasattr(o.index_set(), "is_constructed")
                and o.index_set().is_constructed()
            ):
                for key in sorted(o.keys()):
                    if key == None:
                        break  # Single variable so skip
                    self._add_item(parent=item, o=o[key])

    def _add_item(self, parent, o):
        """
        Add a root item if parent is None, otherwise add a child
        """
        if parent is None:
            item = self._add_root_item(o)
        else:
            item = parent.add_child(o)
        return item

    def _add_root_item(self, o):
        """
        Add a root tree item
        """
        item = ComponentDataItem(None, o, ui_data=self.ui_data)
        self.rootItems.append(item)
        return item

    def parent(self, index):
        if not index.isValid():
            return myqt.QtCore.QModelIndex()
        item = index.internalPointer()
        if item.parent is None:
            return myqt.QtCore.QModelIndex()
        else:
            return self.createIndex(0, 0, item.parent)

    def index(self, row, column, parent=myqt.QtCore.QModelIndex()):
        if not parent.isValid():
            return self.createIndex(row, column, self.rootItems[row])
        parentItem = parent.internalPointer()
        return self.createIndex(row, column, parentItem.children[row])

    def columnCount(self, parent=myqt.QtCore.QModelIndex()):
        """
        Return the number of columns
        """
        return len(self.column)

    def rowCount(self, parent=myqt.QtCore.QModelIndex()):
        if not parent.isValid():
            return len(self.rootItems)
        return len(parent.internalPointer().children)

    def data(
        self, index=myqt.QtCore.QModelIndex(), role=myqt.Qt.ItemDataRole.DisplayRole
    ):
        if (
            role == myqt.Qt.ItemDataRole.DisplayRole
            or role == myqt.Qt.ItemDataRole.EditRole
        ):
            a = self.column[index.column()]
            return index.internalPointer().get(a)
        elif role == myqt.Qt.ItemDataRole.ToolTipRole:
            if self.column[index.column()] == "name":
                o = index.internalPointer()
                if isinstance(o.data, Constraint._ComponentDataClass):
                    return o.get("expr")
                else:
                    return o.get("doc")
        elif role == myqt.Qt.ItemDataRole.ForegroundRole:
            if isinstance(
                index.internalPointer().data, (Block, Block._ComponentDataClass)
            ):
                return myqt.QColor(myqt.QtCore.Qt.GlobalColor.black)
            else:
                return myqt.QColor(myqt.QtCore.Qt.GlobalColor.blue)
        else:
            return

    def headerData(self, i, orientation, role=myqt.Qt.ItemDataRole.DisplayRole):
        """
        Return the column headings for the horizontal header and
        index numbers for the vertical header.
        """
        if (
            orientation == myqt.Qt.Orientation.Horizontal
            and role == myqt.Qt.ItemDataRole.DisplayRole
        ):
            return self.column[i]
        return None

    def flags(self, index=myqt.QtCore.QModelIndex()):
        if self.column[index.column()] in self._col_editable:
            return (
                myqt.Qt.ItemFlag.ItemIsEnabled
                | myqt.Qt.ItemFlag.ItemIsSelectable
                | myqt.Qt.ItemFlag.ItemIsEditable
            )
        else:
            return myqt.Qt.ItemFlag.ItemIsEnabled | myqt.Qt.ItemFlag.ItemIsSelectable



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for expression generation
#

import pickle
import os
import io
import sys
from os.path import abspath, dirname

currdir = dirname(abspath(__file__)) + os.sep

import pyomo.common.unittest as unittest

from pyomo.environ import (
    AbstractModel,
    ConcreteModel,
    Set,
    Var,
    Param,
    Constraint,
    inequality,
    display,
)
from pyomo.core.expr.numvalue import value
from pyomo.core.expr.relational_expr import (
    InequalityExpression,
    EqualityExpression,
    RangedExpression,
)


class TestGenerate_RelationalExpression(unittest.TestCase):
    def setUp(self):
        m = AbstractModel()
        m.I = Set()
        m.a = Var()
        m.b = Var()
        m.c = Var()
        m.x = Var(m.I)
        self.m = m

    def tearDown(self):
        self.m = None

    def test_simpleEquality(self):
        #
        # Check the structure of a simple equality statement
        #
        m = self.m
        e = m.a == m.b
        self.assertIs(type(e), EqualityExpression)
        self.assertEqual(e.nargs(), 2)
        self.assertIs(e.arg(0), m.a)
        self.assertIs(e.arg(1), m.b)

    def test_equalityErrors(self):
        #
        # Check equality errors
        #
        m = self.m
        e = m.a == m.b
        #       =
        #      / \
        #     =   5
        #    / \
        #   a   b
        with self.assertRaisesRegex(
            TypeError,
            "Cannot create an EqualityExpression where one of the "
            "sub-expressions is a relational expression:",
        ):
            e == m.a
        with self.assertRaisesRegex(
            TypeError,
            "Cannot create an EqualityExpression where one of the "
            "sub-expressions is a relational expression:",
        ):
            m.a == e

        #
        # Test expression with an indexed variable
        #
        with self.assertRaisesRegex(
            TypeError, "Argument .* is an indexed numeric value"
        ):
            m.x == m.a
        with self.assertRaisesRegex(
            TypeError, "Argument .* is an indexed numeric value"
        ):
            m.a == m.x

    def test_simpleInequality1(self):
        #
        # Check the structure of a simple inequality
        #
        m = self.m
        #    <
        #   / \
        #  a   b
        e = m.a < m.b
        self.assertIs(type(e), InequalityExpression)
        self.assertEqual(e.nargs(), 2)
        self.assertIs(e.arg(0), m.a)
        self.assertIs(e.arg(1), m.b)
        # self.assertEqual(len(e._strict), 1)
        self.assertEqual(e._strict, True)

        #    <=
        #   / \
        #  a   b
        e = m.a <= m.b
        self.assertIs(type(e), InequalityExpression)
        self.assertEqual(e.nargs(), 2)
        self.assertIs(e.arg(0), m.a)
        self.assertIs(e.arg(1), m.b)
        # self.assertEqual(len(e._strict), 1)
        self.assertEqual(e._strict, False)

        #    >
        #   / \
        #  a   b
        e = m.a > m.b
        self.assertIs(type(e), InequalityExpression)
        self.assertEqual(e.nargs(), 2)
        self.assertIs(e.arg(0), m.b)
        self.assertIs(e.arg(1), m.a)
        # self.assertEqual(len(e._strict), 1)
        self.assertEqual(e._strict, True)

        #    >=
        #   / \
        #  a   b
        e = m.a >= m.b
        self.assertIs(type(e), InequalityExpression)
        self.assertEqual(e.nargs(), 2)
        self.assertIs(e.arg(0), m.b)
        self.assertIs(e.arg(1), m.a)
        # self.assertEqual(len(e._strict), 1)
        self.assertEqual(e._strict, False)

    def test_simpleInequality2(self):
        #
        # Check the structure of a simple inequality
        #
        m = self.m
        #    <
        #   / \
        #  a   b
        e = inequality(lower=m.a, body=m.b, strict=True)
        self.assertIs(type(e), InequalityExpression)
        self.assertEqual(e.nargs(), 2)
        self.assertIs(e.arg(0), m.a)
        self.assertIs(e.arg(1), m.b)
        # self.assertEqual(len(e._strict), 1)
        self.assertEqual(e._strict, True)

        #    <=
        #   / \
        #  a   b
        e = inequality(lower=m.a, upper=m.b)
        self.assertIs(type(e), InequalityExpression)
        self.assertEqual(e.nargs(), 2)
        self.assertIs(e.arg(0), m.a)
        self.assertIs(e.arg(1), m.b)
        # self.assertEqual(len(e._strict), 1)
        self.assertEqual(e._strict, False)

        #    >
        #   / \
        #  a   b
        e = inequality(lower=m.b, upper=m.a, strict=True)
        self.assertIs(type(e), InequalityExpression)
        self.assertEqual(e.nargs(), 2)
        self.assertIs(e.arg(0), m.b)
        self.assertIs(e.arg(1), m.a)
        # self.assertEqual(len(e._strict), 1)
        self.assertEqual(e._strict, True)

        #    >=
        #   / \
        #  a   b
        e = m.a >= m.b
        e = inequality(body=m.b, upper=m.a)
        self.assertIs(type(e), InequalityExpression)
        self.assertEqual(e.nargs(), 2)
        self.assertIs(e.arg(0), m.b)
        self.assertIs(e.arg(1), m.a)
        # self.assertEqual(len(e._strict), 1)
        self.assertEqual(e._strict, False)

        try:
            inequality(None, None)
            self.fail("expected invalid inequality error.")
        except ValueError:
            pass

        try:
            inequality(m.a, None)
            self.fail("expected invalid inequality error.")
        except ValueError:
            pass


class TestGenerate_RangedExpression(unittest.TestCase):
    def setUp(self):
        m = AbstractModel()
        m.I = Set()
        m.a = Var()
        m.b = Var()
        m.c = Var()
        m.x = Var(m.I)
        self.m = m

    def tearDown(self):
        self.m = None

    def test_compoundInequality(self):
        m = self.m
        #       <
        #      / \
        #     <   c
        #    / \
        #   a   b
        e = inequality(m.a, m.b, m.c, strict=True)
        self.assertIs(type(e), RangedExpression)
        self.assertEqual(e.nargs(), 3)
        self.assertIs(e.arg(0), m.a)
        self.assertIs(e.arg(1), m.b)
        self.assertIs(e.arg(2), m.c)
        # self.assertEqual(len(e._strict), 2)
        self.assertEqual(e._strict[0], True)
        self.assertEqual(e._strict[1], True)

        #       <=
        #      / \
        #     <=  c
        #    / \
        #   a   b
        e = inequality(m.a, m.b, m.c)
        self.assertIs(type(e), RangedExpression)
        self.assertEqual(e.nargs(), 3)
        self.assertIs(e.arg(0), m.a)
        self.assertIs(e.arg(1), m.b)
        self.assertIs(e.arg(2), m.c)
        # self.assertEqual(len(e._strict), 2)
        self.assertEqual(e._strict[0], False)
        self.assertEqual(e._strict[1], False)

        #       >
        #      / \
        #     >   c
        #    / \
        #   a   b
        e = inequality(upper=m.c, body=m.b, lower=m.a, strict=True)
        self.assertIs(type(e), RangedExpression)
        self.assertEqual(e.nargs(), 3)
        self.assertIs(e.arg(2), m.c)
        self.assertIs(e.arg(1), m.b)
        self.assertIs(e.arg(0), m.a)
        # self.assertEqual(len(e._strict), 2)
        self.assertEqual(e._strict[0], True)
        self.assertEqual(e._strict[1], True)

        #       >=
        #      / \
        #     >=  c
        #    / \
        #   a   b
        e = inequality(upper=m.c, body=m.b, lower=m.a)
        self.assertIs(type(e), RangedExpression)
        self.assertEqual(e.nargs(), 3)
        self.assertIs(e.arg(2), m.c)
        self.assertIs(e.arg(1), m.b)
        self.assertIs(e.arg(0), m.a)
        # self.assertEqual(len(e._strict), 2)
        self.assertEqual(e._strict[0], False)
        self.assertEqual(e._strict[1], False)

        #       <=
        #      / \
        #     <=  0
        #    / \
        #   0   a
        e = inequality(0, m.a, 0)
        self.assertIs(type(e), RangedExpression)
        self.assertEqual(e.nargs(), 3)
        self.assertIs(e.arg(2), 0)
        self.assertIs(e.arg(1), m.a)
        self.assertIs(e.arg(0), 0)
        # self.assertEqual(len(e._strict), 2)
        self.assertEqual(e._strict[0], False)
        self.assertEqual(e._strict[1], False)

        #       <
        #      / \
        #     <  0
        #    / \
        #   0   a
        e = inequality(0, m.a, 0, True)
        self.assertIs(type(e), RangedExpression)
        self.assertEqual(e.nargs(), 3)
        self.assertIs(e.arg(2), 0)
        self.assertIs(e.arg(1), m.a)
        self.assertIs(e.arg(0), 0)
        # self.assertEqual(len(e._strict), 2)
        self.assertEqual(e._strict[0], True)
        self.assertEqual(e._strict[1], True)

    def test_val1(self):
        m = ConcreteModel()
        m.v = Var(initialize=2)

        e = inequality(0, m.v, 2)
        self.assertEqual(value(e), True)
        e = inequality(0, m.v, 1)
        self.assertEqual(value(e), False)
        e = inequality(0, m.v, 2, strict=True)
        self.assertEqual(value(e), False)

    def test_val2(self):
        m = ConcreteModel()
        m.v = Var(initialize=2)

        e = 1 < m.v
        e = e <= 2
        self.assertEqual(value(e), True)
        e = 1 <= m.v
        e = e < 2
        self.assertEqual(value(e), False)


#
# Fixed               - Expr has a fixed value
# Constant            - Expr only contains constants and immutable parameters
# PotentiallyVariable - Expr contains one or more variables
#
class TestIsFixedIsConstant(unittest.TestCase):
    def setUp(self):
        # This class tests the Pyomo 5.x expression trees

        def d_fn(model):
            return model.c + model.c

        self.model = AbstractModel()
        self.model.a = Var(initialize=1.0)
        self.model.b = Var(initialize=2.0)
        self.model.c = Param(initialize=1, mutable=True)
        self.model.d = Param(initialize=d_fn, mutable=True)
        self.model.e = Param(initialize=d_fn, mutable=False)
        self.model.f = Param(initialize=0, mutable=True)
        self.model.g = Var(initialize=0)
        self.instance = self.model.create_instance()

    def tearDown(self):
        self.model = None
        self.instance = None

    def test_relational_ops(self):
        #
        # Relation of mutable parameters:  fixed, not constant, pvar
        #
        expr = self.instance.c < self.instance.d
        self.assertEqual(expr.is_fixed(), True)
        self.assertEqual(expr.is_constant(), False)
        self.assertEqual(expr.is_potentially_variable(), False)
        expr = inequality(0, self.instance.c, self.instance.d)
        self.assertEqual(expr.is_fixed(), True)
        self.assertEqual(expr.is_constant(), False)
        self.assertEqual(expr.is_potentially_variable(), False)
        expr = self.instance.c == self.instance.d
        self.assertEqual(expr.is_fixed(), True)
        self.assertEqual(expr.is_constant(), False)
        self.assertEqual(expr.is_potentially_variable(), False)
        #
        # Relation of unfixed variable and mutable parameters:  not fixed, not constant, pvar
        #
        expr = self.instance.a <= self.instance.d
        self.assertEqual(expr.is_fixed(), False)
        self.assertEqual(expr.is_constant(), False)
        self.assertEqual(expr.is_potentially_variable(), True)
        expr = inequality(0, self.instance.a, self.instance.d)
        self.assertEqual(expr.is_fixed(), False)
        self.assertEqual(expr.is_constant(), False)
        self.assertEqual(expr.is_potentially_variable(), True)
        expr = self.instance.a == self.instance.d
        self.assertEqual(expr.is_fixed(), False)
        self.assertEqual(expr.is_constant(), False)
        self.assertEqual(expr.is_potentially_variable(), True)
        #
        # Relation of unfixed variables:  not fixed, not constant, pvar
        #
        expr = self.instance.a * self.instance.a >= self.instance.b
        self.assertEqual(expr.is_fixed(), False)
        self.assertEqual(expr.is_constant(), False)
        self.assertEqual(expr.is_potentially_variable(), True)
        expr = self.instance.a * self.instance.a == self.instance.b
        self.assertEqual(expr.is_fixed(), False)
        self.assertEqual(expr.is_constant(), False)
        self.assertEqual(expr.is_potentially_variable(), True)
        expr = inequality(self.instance.b, self.instance.a * self.instance.a, 0)
        self.assertEqual(expr.is_fixed(), False)
        self.assertEqual(expr.is_constant(), False)
        self.assertEqual(expr.is_potentially_variable(), True)
        #
        # Relation of fixed and unfixed variables:  not fixed, not constant, pvar
        #
        self.instance.a.fixed = True
        self.assertEqual(expr.is_fixed(), False)
        self.assertEqual(expr.is_constant(), False)
        self.assertEqual(expr.is_potentially_variable(), True)
        #
        # Relation of fixed variables:  fixed, not constant, pvar
        #
        self.instance.b.fixed = True
        self.assertEqual(expr.is_fixed(), True)
        self.assertEqual(expr.is_constant(), False)
        self.assertEqual(expr.is_potentially_variable(), True)


class TestMultiArgumentExpressions(unittest.TestCase):
    def test_double_sided_ineq(self):
        m = ConcreteModel()
        m.s = Set(initialize=[1.0, 2.0, 3.0, 4.0, 5.0])

        m.vmin = Param(m.s, initialize=lambda m, i: i)
        m.vmax = Param(m.s, initialize=lambda m, i: i**2)

        m.v = Var(m.s)

        def _con(m, i):
            return inequality(m.vmin[i] ** 2, m.v[i], m.vmax[i] ** 2)

        m.con = Constraint(m.s, rule=_con)

        OUT = io.StringIO()
        for i in m.s:
            OUT.write(str(_con(m, i)))
            OUT.write("\n")
        display(m.con, ostream=OUT)

        reference = """1.0  <=  v[1.0]  <=  1.0
4.0  <=  v[2.0]  <=  16.0
9.0  <=  v[3.0]  <=  81.0
16.0  <=  v[4.0]  <=  256.0
25.0  <=  v[5.0]  <=  625.0
con : Size=5
    Key : Lower : Body : Upper
    1.0 :   1.0 : None :   1.0
    2.0 :   4.0 : None :  16.0
    3.0 :   9.0 : None :  81.0
    4.0 :  16.0 : None : 256.0
    5.0 :  25.0 : None : 625.0
"""
        self.assertEqual(OUT.getvalue(), reference)


#
# Test pickle logic
#
class Test_pickle(unittest.TestCase):
    def test_ineq(self):
        M = ConcreteModel()
        M.v = Var()
        e = M.v >= 0
        s = pickle.dumps(e)
        e_ = pickle.loads(s)
        self.assertEqual(str(e), str(e_))

    def test_range(self):
        M = ConcreteModel()
        M.v = Var()
        e = inequality(0, M.v, 1)
        s = pickle.dumps(e)
        e_ = pickle.loads(s)
        self.assertEqual(str(e), str(e_))

    def test_eq(self):
        M = ConcreteModel()
        M.v = Var()
        e = M.v == 0
        s = pickle.dumps(e)
        e_ = pickle.loads(s)
        self.assertEqual(str(e), str(e_))


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
import operator
from math import prod as _prod

import pyomo.core.expr as EXPR
from pyomo.common import DeveloperError
from pyomo.common.collections import ComponentMap
from pyomo.common.dependencies import attempt_import
from pyomo.common.errors import NondifferentiableError
from pyomo.core.expr.numvalue import value, native_types

#
# Sympy takes a significant time to load; defer importing it unless
# someone actually needs the interface.
#

_operatorMap = {}
_pyomo_operator_map = {}
_functionMap = {}


def _nondifferentiable(x):
    if type(x[1]) is tuple:
        # sympy >= 1.3 returns tuples (var, order)
        wrt = x[1][0]
    else:
        # early versions of sympy returned the bare var
        wrt = x[1]
    raise NondifferentiableError(
        "The sub-expression '%s' is not differentiable with respect to %s" % (x[0], wrt)
    )


def _external_fcn(*x):
    raise TypeError(
        "Expressions containing external functions are not convertible to "
        f"sympy expressions (found 'f{x}')"
    )


def _configure_sympy(sympy, available):
    if not available:
        return

    _operatorMap.update(
        {
            sympy.Add: sum,
            sympy.Mul: _prod,
            sympy.Pow: lambda x: operator.pow(*x),
            sympy.exp: lambda x: EXPR.exp(*x),
            sympy.log: lambda x: EXPR.log(*x),
            sympy.sin: lambda x: EXPR.sin(*x),
            sympy.asin: lambda x: EXPR.asin(*x),
            sympy.sinh: lambda x: EXPR.sinh(*x),
            sympy.asinh: lambda x: EXPR.asinh(*x),
            sympy.cos: lambda x: EXPR.cos(*x),
            sympy.acos: lambda x: EXPR.acos(*x),
            sympy.cosh: lambda x: EXPR.cosh(*x),
            sympy.acosh: lambda x: EXPR.acosh(*x),
            sympy.tan: lambda x: EXPR.tan(*x),
            sympy.atan: lambda x: EXPR.atan(*x),
            sympy.tanh: lambda x: EXPR.tanh(*x),
            sympy.atanh: lambda x: EXPR.atanh(*x),
            sympy.ceiling: lambda x: EXPR.ceil(*x),
            sympy.floor: lambda x: EXPR.floor(*x),
            sympy.sqrt: lambda x: EXPR.sqrt(*x),
            sympy.Abs: lambda x: abs(*x),
            sympy.Derivative: _nondifferentiable,
            sympy.Tuple: lambda x: x,
            sympy.Or: lambda x: EXPR.lor(*x),
            sympy.And: lambda x: EXPR.land(*x),
            sympy.Implies: lambda x: EXPR.implies(*x),
            sympy.Equivalent: lambda x: EXPR.equivalents(*x),
            sympy.Not: lambda x: EXPR.lnot(*x),
            sympy.LessThan: lambda x: operator.le(*x),
            sympy.StrictLessThan: lambda x: operator.lt(*x),
            sympy.GreaterThan: lambda x: operator.ge(*x),
            sympy.StrictGreaterThan: lambda x: operator.gt(*x),
            sympy.Equality: lambda x: operator.eq(*x),
        }
    )

    _pyomo_operator_map.update(
        {
            EXPR.SumExpression: sympy.Add,
            EXPR.LinearExpression: sympy.Add,
            EXPR.ProductExpression: sympy.Mul,
            EXPR.MonomialTermExpression: sympy.Mul,
            EXPR.ExternalFunctionExpression: _external_fcn,
            EXPR.AndExpression: sympy.And,
            EXPR.OrExpression: sympy.Or,
            EXPR.ImplicationExpression: sympy.Implies,
            EXPR.EquivalenceExpression: sympy.Equivalent,
            EXPR.XorExpression: sympy.Xor,
            EXPR.NotExpression: sympy.Not,
        }
    )

    _functionMap.update(
        {
            'exp': sympy.exp,
            'log': sympy.log,
            'log10': lambda x: sympy.log(x) / sympy.log(10),
            'sin': sympy.sin,
            'asin': sympy.asin,
            'sinh': sympy.sinh,
            'asinh': sympy.asinh,
            'cos': sympy.cos,
            'acos': sympy.acos,
            'cosh': sympy.cosh,
            'acosh': sympy.acosh,
            'tan': sympy.tan,
            'atan': sympy.atan,
            'tanh': sympy.tanh,
            'atanh': sympy.atanh,
            'ceil': sympy.ceiling,
            'floor': sympy.floor,
            'sqrt': sympy.sqrt,
        }
    )


sympy, sympy_available = attempt_import('sympy', callback=_configure_sympy)


class PyomoSympyBimap:
    def __init__(self):
        self.pyomo2sympy = ComponentMap()
        self.sympy2pyomo = {}
        self.i = 0

    def getPyomoSymbol(self, sympy_object, default=None):
        return self.sympy2pyomo.get(sympy_object, default)

    def getSympySymbol(self, pyomo_object):
        if pyomo_object in self.pyomo2sympy:
            return self.pyomo2sympy[pyomo_object]
        # Pyomo currently ONLY supports Real variables (not complex
        # variables).  If that ever changes, then we will need to
        # revisit hard-coding the symbol type here
        sympy_obj = sympy.Symbol("x%d" % self.i, real=True)
        self.i += 1
        self.pyomo2sympy[pyomo_object] = sympy_obj
        self.sympy2pyomo[sympy_obj] = pyomo_object
        return sympy_obj

    def sympyVars(self):
        return self.sympy2pyomo.keys()


# =====================================================
# sympyify_expression
# =====================================================


class Pyomo2SympyVisitor(EXPR.StreamBasedExpressionVisitor):
    def __init__(self, object_map, keep_mutable_parameters=False):
        sympy.Add  # this ensures _configure_sympy gets run
        super(Pyomo2SympyVisitor, self).__init__()
        self.object_map = object_map
        self.keep_mutable_parameters = keep_mutable_parameters

    def initializeWalker(self, expr):
        return self.beforeChild(None, expr, None)

    def exitNode(self, node, values):
        if node.__class__ is EXPR.UnaryFunctionExpression:
            return _functionMap[node._name](values[0])
        _op = _pyomo_operator_map.get(node.__class__, None)
        if _op is None:
            return node._apply_operation(values)
        else:
            return _op(*tuple(values))

    def beforeChild(self, node, child, child_idx):
        #
        # Don't replace native or sympy types
        #
        if type(child) in native_types:
            return False, child
        #
        # Replace pyomo variables with sympy variables
        #
        if child.is_potentially_variable():
            #
            # We will descend into all expressions...
            #
            if child.is_expression_type():
                return True, None
            else:
                return False, self.object_map.getSympySymbol(child)
        #
        # Everything else is a constant...
        #
        if self.keep_mutable_parameters and child.is_parameter_type() and child.mutable:
            return False, self.object_map.getSympySymbol(child)
        return False, value(child)


class Sympy2PyomoVisitor(EXPR.StreamBasedExpressionVisitor):
    def __init__(self, object_map):
        sympy.Add  # this ensures _configure_sympy gets run
        super(Sympy2PyomoVisitor, self).__init__()
        self.object_map = object_map

    def initializeWalker(self, expr):
        return self.beforeChild(None, expr, None)

    def enterNode(self, node):
        return (node.args, [])

    def exitNode(self, node, values):
        """Visit nodes that have been expanded"""
        _op = _operatorMap.get(node.func, None)
        if _op is None:
            raise DeveloperError(
                f"sympy expression type {node.func} not found in the operator map"
            )
        return _op(tuple(values))

    def beforeChild(self, node, child, child_idx):
        if not child.args:
            item = self.object_map.getPyomoSymbol(child, None)
            if item is None:
                item = float(child.evalf())
            return False, item
        return True, None


def sympyify_expression(expr, keep_mutable_parameters=False):
    """Convert a Pyomo expression to a Sympy expression"""
    #
    # Create the visitor and call it.
    #
    object_map = PyomoSympyBimap()
    visitor = Pyomo2SympyVisitor(
        object_map, keep_mutable_parameters=keep_mutable_parameters
    )
    return object_map, visitor.walk_expression(expr)


def sympy2pyomo_expression(expr, object_map):
    visitor = Sympy2PyomoVisitor(object_map)
    return visitor.walk_expression(expr)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from collections.abc import MutableSet

from pyomo.common.autoslots import AutoSlots


class OrderedSet(AutoSlots.Mixin, MutableSet):
    __slots__ = ("_dict",)

    def __init__(self, iterable=None):
        # Starting in Python 3.7, dict is ordered (and is faster than
        # OrderedDict).  dict began supporting reversed() in 3.8.
        self._dict = {}
        if iterable is not None:
            self.update(iterable)

    def __str__(self):
        """String representation of the mapping."""
        return "OrderedSet(%s)" % (", ".join(repr(x) for x in self))

    def update(self, iterable):
        if isinstance(iterable, OrderedSet):
            self._dict.update(iterable._dict)
        else:
            self._dict.update((val, None) for val in iterable)

    #
    # Implement MutableSet abstract methods
    #

    def __contains__(self, val):
        return val in self._dict

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)

    def add(self, val):
        """Add an element."""
        self._dict[val] = None

    def discard(self, val):
        """Remove an element. Do not raise an exception if absent."""
        if val in self._dict:
            del self._dict[val]

    #
    # The remaining MutableSet methods have slow default
    # implementations.
    #

    def clear(self):
        """Remove all elements from this set."""
        self._dict.clear()

    def remove(self, val):
        """Remove an element. If not a member, raise a KeyError."""
        del self._dict[val]

    def intersection(self, other):
        other = set(other)
        res = OrderedSet(filter(other.__contains__, self))
        return res

    def union(self, other):
        res = OrderedSet(self)
        res.update(other)
        return res

    #
    # Not strictly part of MutableSet, but it makes sense that OrderedSet
    # should be reversible
    #
    def __reversed__(self):
        return reversed(self._dict)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.extensions import ExtensionBuilderFactory
from .base import SolverFactory
from .solvers import Gurobi, Ipopt, Cbc, Cplex, Highs, MAiNGO
from .build import AppsiBuilder


def load():
    ExtensionBuilderFactory.register('appsi')(AppsiBuilder)
    SolverFactory.register(
        name='gurobi', doc='Automated persistent interface to Gurobi'
    )(Gurobi)
    SolverFactory.register(name='cplex', doc='Automated persistent interface to Cplex')(
        Cplex
    )
    SolverFactory.register(name='ipopt', doc='Automated persistent interface to Ipopt')(
        Ipopt
    )
    SolverFactory.register(name='cbc', doc='Automated persistent interface to Cbc')(Cbc)
    SolverFactory.register(name='highs', doc='Automated persistent interface to Highs')(
        Highs
    )
    SolverFactory.register(
        name='maingo', doc='Automated persistent interface to MAiNGO'
    )(MAiNGO)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


class SymbolMap:
    """
    A class for tracking assigned labels for modeling components.

    Symbol maps are used, for example, when writing problem files for
    input to an optimizer.

    Warning:
        A symbol map should never be pickled.  This class is
        typically constructed by solvers and writers, and it may be
        owned by models.

    Note:
        We should change the API to not use camelcase.

    Attributes
    ----------
    byObject : dict
        maps (object id) to (string label)
    bySymbol : dict
        maps (string label) to (object)
    aliases : dict
        maps (string label) to (object)
    default_labeler:
        used to compute a string label from an object
    """

    def __init__(self, labeler=None):
        self.byObject = {}
        self.bySymbol = {}
        self.aliases = {}
        self.default_labeler = labeler

    class UnknownSymbol:
        pass

    def __getstate__(self):
        #
        # While we should generally not pickle a SymbolMap, we still
        # need to implement __getstate__ / __setstate__ so that the
        # bi-map is correctly duplicated if the object is ever
        # deepcopied (the id() keys need to be updated to point to the
        # new model objects)
        #
        # Note: byObject and bySymbol constitute a bimap.  We only need
        # to save one of them.
        #
        return (self.bySymbol, self.aliases, self.default_labeler)

    def __setstate__(self, state):
        self.bySymbol, self.aliases, self.default_labeler = state
        self.byObject = {id(v): k for k, v in self.bySymbol.items()}

    def addSymbol(self, obj, symb):
        """
        Add a symbol for a given object

        This method assumes that objects and symbol names will not conflict.
        """
        nSymbols = len(self.byObject) + 1
        self.byObject[id(obj)] = symb
        self.bySymbol[symb] = obj
        if nSymbols != len(self.bySymbol):
            raise RuntimeError(
                "SymbolMap.addSymbol(): duplicate symbol.  "
                "SymbolMap likely in an inconsistent state"
            )
        if len(self.byObject) != len(self.bySymbol):
            raise RuntimeError(
                "SymbolMap.addSymbol(): duplicate object.  "
                "SymbolMap likely in an inconsistent state"
            )

    def addSymbols(self, obj_symbol_tuples):
        """
        Add (object, symbol) tuples from an iterable object.

        This method assumes that objects and symbol names will not conflict.
        """
        nSymbols = len(self.bySymbol)
        for obj, symbol in obj_symbol_tuples:
            self.byObject[id(obj)] = symbol
            self.bySymbol[symbol] = obj
            nSymbols += 1
        if nSymbols != len(self.bySymbol):
            raise RuntimeError(
                "SymbolMap.addSymbols(): duplicate symbol.  "
                "SymbolMap likely in an inconsistent state"
            )
        if len(self.byObject) != len(self.bySymbol):
            raise RuntimeError(
                "SymbolMap.addSymbols(): duplicate object.  "
                "SymbolMap likely in an inconsistent state"
            )

    def createSymbol(self, obj, labeler=None, *args):
        """
        Create a symbol for an object with a given labeler.  No
        error checking is done to ensure that the generated symbol
        name is unique.
        """
        if labeler is None:
            if self.default_labeler is not None:
                labeler = self.default_labeler
            else:
                labeler = str
        symbol = labeler(obj, *args)
        self.addSymbol(obj, symbol)
        return symbol

    def createSymbols(self, objs, labeler=None, *args):
        """
        Create a symbol for iterable objects with a given labeler.  No
        error checking is done to ensure that the generated symbol
        names are unique.
        """
        if labeler is None:
            if self.default_labeler is not None:
                labeler = self.default_labeler
            else:
                labeler = str
        self.addSymbols((obj, labeler(obj, *args)) for obj in objs)

    def getSymbol(self, obj, labeler=None, *args):
        """
        Return the symbol for an object.  If it has not already been cached
        in the symbol map, then create it.
        """
        obj_id = id(obj)
        if obj_id in self.byObject:
            return self.byObject[obj_id]
        #
        # Create a new symbol, performing an error check if it is a duplicate
        #
        symbol = (labeler or self.default_labeler or str)(obj, *args)
        if symbol in self.bySymbol:
            # The labeler can have side-effects, including registering
            # this symbol in the symbol map
            if obj is self.bySymbol[symbol]:
                return symbol
            raise RuntimeError(
                "Duplicate symbol '%s' already associated with "
                "component '%s' (conflicting component: '%s')"
                % (symbol, self.bySymbol[symbol].name, obj.name)
            )
        self.bySymbol[symbol] = obj
        self.byObject[obj_id] = symbol
        return symbol

    def alias(self, obj, name):
        """
        Create an alias for an object.  An aliases are symbols that
        do not have a one-to-one correspondence with objects.
        """
        if name in self.aliases:
            #
            # If the alias exists and the objects are the same,
            # then return.  Otherwise, raise an exception.
            #
            old_object = self.aliases[name]
            if old_object is obj:
                return
            else:
                raise RuntimeError(
                    "Duplicate alias '%s' already associated with "
                    "component '%s' (conflicting component: '%s')"
                    % (
                        name,
                        "UNKNOWN" if old_object is None else old_object.name,
                        obj.name,
                    )
                )
        else:
            #
            # Add the alias
            #
            self.aliases[name] = obj

    def getObject(self, symbol):
        """
        Return the object corresponding to a symbol
        """
        if symbol in self.bySymbol:
            return self.bySymbol[symbol]
        elif symbol in self.aliases:
            return self.aliases[symbol]
        else:
            return SymbolMap.UnknownSymbol

    def removeSymbol(self, obj):
        symb = self.byObject.pop(id(obj))
        self.bySymbol.pop(symb)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Attributes describing the current platform and user configuration.

This module provides standardized attributes that other parts of Pyomo
can use to interrogate aspects of the current platform, and to find
information about the current user configuration (including where to
locate the main Pyomo configuration directory).

"""

import os
import platform

_platform = platform.system().lower()
#: bool : True if running in a "native" Windows environment
is_native_windows = _platform.startswith('windows')
#: bool : True if running on Windows (native or cygwin)
is_windows = is_native_windows or _platform.startswith('cygwin')
#: bool : True if running on Mac/OSX
is_osx = _platform.startswith('darwin')
#: bool: True if running under the PyPy interpreter
is_pypy = platform.python_implementation().lower().startswith('pypy')

#: str : Absolute path to the user's Pyomo Configuration Directory.
#:
#:     By default, this is ``~/.pyomo`` on Linux and OSX and
#:     ``%LOCALAPPDATA%/Pyomo`` on Windows.  It can be overridden by
#:     setting the ``PYOMO_CONFIG_DIR`` environment variable before
#:     importing Pyomo.
PYOMO_CONFIG_DIR = None

if 'PYOMO_CONFIG_DIR' in os.environ:
    PYOMO_CONFIG_DIR = os.path.abspath(os.environ['PYOMO_CONFIG_DIR'])
elif is_windows:
    PYOMO_CONFIG_DIR = os.path.abspath(
        os.path.join(os.environ.get('LOCALAPPDATA', ''), 'Pyomo')
    )
else:
    PYOMO_CONFIG_DIR = os.path.abspath(
        os.path.join(os.environ.get('HOME', ''), '.pyomo')
    )

# Note that alternative platform-independent implementation of the above
# could be to use:
#
#   PYOMO_CONFIG_DIR = os.path.abspath(appdirs.user_data_dir('pyomo'))
#
# But would require re-adding the hard dependency on appdirs.  For now
# (13 Jul 20), the above appears to be sufficiently robust.
